#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Unit tests for schooltool.relationship
"""

import unittest
import doctest

from schooltool.relationship.tests import URIStub


def doctest_relate():
    """Tests for relate

    getRelatedObjects relies on adapters to IRelationshipLinks.  For the
    purposes of this test it is simpler to just implement IRelationshipLinks
    directly in the object

        >>> from zope.interface import implements
        >>> from schooltool.relationship.interfaces import IRelationshipLinks

        >>> class Relatable:
        ...     implements(IRelationshipLinks)
        ...     def __init__(self, name):
        ...         self._name = name
        ...     def __repr__(self):
        ...         return self._name
        ...     def __iter__(self):
        ...         return iter([])
        ...     def add(self, link):
        ...         print 'Linking %s with %s (the %s in %s)' % (self,
        ...                 link.target, link.role.uri, link.rel_type)

        >>> fred = Relatable('Fred')
        >>> wilma = Relatable('Wilma')

    Now we can test relate

        >>> from schooltool.relationship.relationship import relate

        >>> husband = URIStub('husband')
        >>> wife = URIStub('wife')

        >>> relate('marriage', (fred, husband), (wilma, wife))
        Linking Fred with Wilma (the wife in marriage)
        Linking Wilma with Fred (the husband in marriage)

    """


def doctest_LinkSet_getTargetsByRole():
    """Tests for getTargetsByRole

        >>> from schooltool.relationship.relationship import LinkSet, Link

        >>> obj = LinkSet()

        >>> role_a = URIStub('role_of_a')
        >>> role_b = URIStub('role_of_b')

        >>> obj.add(Link(role_b, 'a', role_a, 'rel_type_a'))
        >>> obj.add(Link(role_a, 'b', role_b, 'rel_type_b'))

    Now we can test getTargetsByRole

        >>> obj.getTargetsByRole(role_a)
        ['a']
        >>> obj.getTargetsByRole(role_b)
        ['b']
        >>> obj.getTargetsByRole(URIStub('role_c'))
        []

    """


def doctest_RelationshipSchema():
    """Tests for RelationshipSchema

        >>> from schooltool.relationship.tests import setUp, tearDown
        >>> setUp()

    The constructor takes exactly two keyword arguments

        >>> from schooltool.relationship import RelationshipSchema
        >>> role_mgr = URIStub('example:Mgr')
        >>> role_rpt = URIStub('example:Rpt')
        >>> role_spv = URIStub('example:Spv')
        >>> RelationshipSchema('example:Mgmt', manager=role_mgr,
        ...                    report=role_rpt, supervisor=role_spv)
        Traceback (most recent call last):
          ...
        TypeError: A relationship must have exactly two ends.
        >>> RelationshipSchema('example:Mgmt', manager='example:Mgr')
        Traceback (most recent call last):
          ...
        TypeError: A relationship must have exactly two ends.

    This works:

        >>> Management = RelationshipSchema('example:Mgmt',
        ...                                 manager=role_mgr,
        ...                                 report=role_rpt)

    You can call relationship schemas

        >>> from schooltool.relationship.tests import SomeObject
        >>> a = SomeObject('a')
        >>> b = SomeObject('b')
        >>> Management(manager=a, report=b)

    You will see that a is b's manager, and b is a's report:

        >>> from schooltool.relationship import getRelatedObjects
        >>> getRelatedObjects(b, role_mgr)
        [a]
        >>> getRelatedObjects(a, role_rpt)
        [b]

    Order of arguments does not matter

        >>> c, d = map(SomeObject, ['c', 'd'])
        >>> Management(report=c, manager=d)
        >>> getRelatedObjects(c, role_mgr)
        [d]
        >>> getRelatedObjects(d, role_rpt)
        [c]

    You must give correct arguments, though

        >>> Management(report=c, friend=d)
        Traceback (most recent call last):
          ...
        TypeError: Missing a 'manager' keyword argument.

        >>> Management(manager=c, friend=d)
        Traceback (most recent call last):
          ...
        TypeError: Missing a 'report' keyword argument.

    You should not give extra arguments either

        >>> Management(report=c, manager=b, friend=d)
        Traceback (most recent call last):
          ...
        TypeError: Too many keyword arguments.

    Cleanup

        >>> tearDown()

    """


def doctest_unrelateAll():
    r"""Tests for unrelateAll.

        >>> from schooltool.relationship.tests import setUp, tearDown
        >>> setUp()

    Let us catch all events and remember them

        >>> events = []
        >>> import zope.event
        >>> old_subscribers = zope.event.subscribers[:]
        >>> zope.event.subscribers.append(events.append)

    Nothing happens if the object has no relationships.

        >>> from schooltool.relationship.tests import SomeObject
        >>> a = SomeObject('a')
        >>> from schooltool.relationship import unrelateAll
        >>> unrelateAll(a)
        >>> events
        []

    Suppose that the object has a number of relationships

        >>> from schooltool.relationship import relate
        >>> b, c, d = map(SomeObject, ['b', 'c', 'd'])
        >>> relationships = [
        ...       ('example:SomeRelationship', (a, 'example:Foo'),
        ...                                    (b, 'example:Bar')),
        ...       ('example:SomeRelationship', (a, 'example:Foo'),
        ...                                    (c, 'example:Bar')),
        ...       ('example:OtherRelationship', (a, 'example:Symmetric'),
        ...                                     (d, 'example:Symmetric')),
        ...       ('example:Loop', (a, 'example:OneEnd'),
        ...                        (a, 'example:OtherEnd')),
        ...       ('example:Loop', (a, 'example:BothEnds'),
        ...                        (a, 'example:BothEnds')),
        ... ]
        >>> for rel_type, (a, rel_a), (b, rel_b) in relationships:
        ...     relate(rel_type, (a, URIStub(rel_a)), (b, URIStub(rel_b)))

    We are not interested in relationship events up to this point

        >>> del events[:]

    We call `unrelateAll` and it suddenly has no relationships

        >>> unrelateAll(a)

        >>> from schooltool.relationship.interfaces import IRelationshipLinks
        >>> list(IRelationshipLinks(a))
        []

    Relationships are broken properly, from both ends

        >>> from schooltool.relationship import getRelatedObjects
        >>> getRelatedObjects(b, URIStub('example:Foo'))
        []

    Also, we got a bunch of events

        >>> from schooltool.relationship.interfaces \
        ...         import IBeforeRemovingRelationshipEvent
        >>> from schooltool.relationship.interfaces \
        ...         import IRelationshipRemovedEvent
        >>> before_removal_events = set([
        ...         (e.rel_type, (e.participant1, e.role1.uri),
        ...                      (e.participant2, e.role2.uri))
        ...         for e in events
        ...         if IBeforeRemovingRelationshipEvent.providedBy(e)])
        >>> before_removal_events == set(relationships)
        True

        >>> removal_events = set([(e.rel_type,
        ...                        (e.participant1, e.role1.uri),
        ...                        (e.participant2, e.role2.uri))
        ...                       for e in events
        ...                       if IRelationshipRemovedEvent.providedBy(e)])
        >>> removal_events == set(relationships)
        True

        >>> zope.event.subscribers[:] = old_subscribers
        >>> tearDown()

    """


def doctest_BoundRelationshipProperty():
    """Tests for BoundRelationshipProperty.

        >>> from schooltool.relationship.tests import setUp, tearDown
        >>> setUp()

    Set up two types of membership.

        >>> role_student = URIStub('example:Student')
        >>> role_instructor = URIStub('example:Instructor')
        >>> role_course = URIStub('example:Course')

        >>> uri_attending = URIStub('example:Attending')
        >>> uri_instruction = URIStub('example:Instruction')

        >>> from schooltool.relationship import RelationshipSchema
        >>> Instruction = RelationshipSchema(uri_instruction,
        ...                                  instructor=role_instructor,
        ...                                  course=role_course)
        >>> Attending = RelationshipSchema(uri_attending,
        ...                                student=role_student,
        ...                                course=role_course)

    Create Course and Person classes.  We will use the RelationshipProperty
    that should bind to an instance as BoundRelationshipProperty.

        >>> from schooltool.relationship.tests import SomeObject
        >>> from schooltool.relationship.relationship import RelationshipProperty

        >>> class Course(SomeObject):
        ...     students = RelationshipProperty(
        ...         uri_attending, role_course, role_student)
        ...     instructors = RelationshipProperty(
        ...         uri_instruction, role_course, role_instructor)

        >>> class Person(SomeObject):
        ...     attends = RelationshipProperty(
        ...         uri_attending, role_student, role_course)
        ...     instructs = RelationshipProperty(
        ...         uri_instruction, role_instructor, role_course)

    Set up a course with several students.

        >>> course_a = Course('course A')

        >>> john, peter, cathy = students = [
        ...     Person(name) for name in ['John', 'Peter', 'Cathy']]
        >>> for student in students:
        ...     Attending(student=student, course=course_a)

    Set up a teacher that instructs the two courses.

        >>> teacher = Person('William')
        >>> course_b = Course('course B')
        >>> Instruction(instructor=teacher, course=course_a)
        >>> Instruction(instructor=teacher, course=course_b)

    We're done with preparations.

    Check that relationship properties were bound.

        >>> course_a.students
        <schooltool.relationship.relationship.BoundRelationshipProperty object ...>
        >>> teacher.instructs
        <schooltool.relationship.relationship.BoundRelationshipProperty object ...>

    They can be used to add and remove relationships, can be iterated to obtain
    related objects and have other useful methods.

        >>> bool(course_b.students)
        False

        >>> for student in students:
        ...     course_b.students.add(student)

        >>> len(course_b.students)
        3

        >>> course_b.students.remove(john)

        >>> len(course_b.students)
        2

        >>> bool(course_b.students)
        True

        >>> list(course_b.students)
        [Peter, Cathy]

    You can also obtain RelationshipInfo helpers for related objects.

        >>> course_a.instructors.relationships
        [<schooltool.relationship.relationship.RelationshipInfo object ...>]

        >>> rel_info = course_a.instructors.relationships[0]
        >>> rel_info.source
        course A
        >>> rel_info.target
        William

    Notice that 'source' in RelationshipInfo is the instace that
    BoundRelationshipProperty is bound to.  Let's look at the info for a class
    that William teaches.

        >>> list(teacher.instructs)
        [course A, course B]

        >>> rel_info = teacher.instructs.relationships[0]
        >>> rel_info.source
        William
        >>> rel_info.target
        course A

    Finally, let's check that BoundRelationshipProperty.relationships are
    filtered correctly.

        >>> [info.target for info in course_b.students.relationships]
        [Peter, Cathy]
        >>> [info.target for info in course_b.instructors.relationships]
        [William]

        >>> tearDown()

    """


def test_suite():
    optionflags = doctest.NORMALIZE_WHITESPACE | doctest.ELLIPSIS
    return unittest.TestSuite([
                doctest.DocFileSuite('../README.txt'),
                doctest.DocTestSuite('schooltool.relationship.relationship'),
                doctest.DocTestSuite(optionflags=optionflags),
           ])

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Testing registries.
"""
__docformat__ = 'restructuredtext'


_registries = {}

def register(reg_name, func, *args, **kw):
    reg = _registries.setdefault(reg_name, [])
    if reg == []:
        def setupUsingRegistry(*args, **kw):
            setup(reg_name, *args, **kw)
        globals()['setup'+reg_name] = setupUsingRegistry

    reg.append((func, args, kw))


def setup(reg_name, *args, **kw):
    for func, orig_args, orig_kw in _registries[reg_name]:
        all_kw = orig_kw.copy()
        all_kw.update(kw)
        func(*(orig_args+args), **all_kw)



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
SchoolTool application interfaces
"""

import zope.schema

from zope.component.interfaces import ObjectEvent, IObjectEvent
from zope.container.interfaces import IContainer
from zope.container.interfaces import IReadContainer
from zope.container.constraints import contains
from zope.interface import implements
from zope.interface import Interface, Attribute
from zope.location.interfaces import IContained
from zope.authentication.interfaces import IAuthentication, ILogout

from schooltool.app.utils import vocabulary
from schooltool.common.fields import Image
from schooltool.common import SchoolToolMessage as _
from schooltool.person.interfaces import ICalendarDisplayPreferences

# XXX: misplaced calendar integration
from schooltool.calendar.interfaces import ICalendarEvent
from schooltool.calendar.interfaces import ISchoolToolCalendarEvent
from schooltool.calendar.interfaces import ISchoolToolCalendar
from schooltool.calendar.interfaces import IWriteCalendar
from schooltool.calendar.interfaces import IHaveCalendar

import zope.formlib.textwidgets
zope.formlib.textwidgets._ = _
# Here we do a particulary evil thing: we override the translation (_) function
# in the textwidgets module.  This means that all the messages in that module
# are now in the 'schooltool' domain.  This is the list of the messages
# (don't remove the list, it is used in localizable string extraction).
textwidgets_strings=[_('Form input is not a file object'),
                     _("Invalid integer data"),
                     _("Invalid text data"),
                     _("Invalid textual data"),
                     _("Invalid unicode data"),
                     _("Invalid integer data"),
                     _("Invalid floating point data"),
                     _("Invalid datetime data")]


# Events

class IApplicationInitializationEvent(IObjectEvent):
    """The SchoolTool application is being initialized.

    Usually subscribers add something to the initialization process.
    """

class ApplicationInitializationEvent(ObjectEvent):
    implements(IApplicationInitializationEvent)


class IApplicationStartUpEvent(IObjectEvent):
    """The SchoolTool application has started up.

    A hook for plugin initialization.
    """

class ApplicationStartUpEvent(ObjectEvent):
    implements(IApplicationStartUpEvent)


class ICatalogSetUpEvent(IObjectEvent):
    """Old style notification that the SchoolTool catalogs are being initialized.

    Subject to deprecation.  Subscribers should no longer use this to set up their
    catalogs.
    """

class CatalogSetUpEvent(ObjectEvent):
    implements(ICatalogSetUpEvent)


class ICatalogStartUpEvent(IObjectEvent):
    """The SchoolTool catalogs has started up.

    A hook for catalog initialization.
    """

class CatalogStartUpEvent(ObjectEvent):
    implements(ICatalogStartUpEvent)


class ISchoolToolApplication(IReadContainer):
    """The main SchoolTool application object.

    The application is a read-only container with the following items:

        'persons' - IPersonContainer
        'groups' - IGroupContainer
        'resources' - IResourceContainer

    This object can be added as a regular content object to a folder, or
    it can be used as the application root object.
    """

    title = zope.schema.TextLine(
        title=_("Title"),
        required=True,
        description=_(
            "The name for the school or organization running "
            "this server.  This will be displayed on the public calendar, the "
            "bottom of all pages and in the page title."))


class IApplicationPreferences(ICalendarDisplayPreferences):
    """Preferences stored in an annotation on the SchoolToolApplication."""

    title = zope.schema.TextLine(
        title=_("Title"),
        required=True,
        description=_(
            "The name for the school or organization running "
            "this server.  This will be displayed on the public calendar, the "
            "bottom of all pages and in the page title."))

    frontPageCalendar = zope.schema.Bool(
        title=_("Front Page Calendar"),
        description=_(
            "Display site-wide calendar as the front page of the site."),
        required=False,
        default=True)

    logo = Image(
        title=_("School logo"),
        format="JPEG",
        required=False)

    name_sorting = zope.schema.Choice(
        title=_('Name sorting'),
        vocabulary=vocabulary([('first_name', _('First name')),
                               ('last_name', _('Last name'))]))


class IApplicationTabs(Interface):
    """Tab visible settings stored in an annotation on SchoolToolApplication."""


class ISchoolToolAuthentication(IAuthentication, ILogout):
    """A local authentication utility for SchoolTool"""

    def setCredentials(request, username, password):
        """Save the username and password in the session.

        If the credentials passed are invalid, ValueError is raised.
        """

    def clearCredentials(request):
        """Forget the username and password stored in a session"""


class ICalendarParentCrowd(Interface):
    """A crowd object that is used on a calendar's parent.

    This is just a marker interface.
    """

    def contains(principal):
        """Return True if principal is in the crowd."""


class IAsset(Interface):
    """An asset of a leader."""

    leaders = Attribute("Leaders of this asset")


class ICookieLanguageSelector(Interface):

    def getLanguageList():
        """Return the list of available languages."""

    def getSelectedLanguage():
        """Return the language that is currently selected."""

    def setSelectedLanguage():
        """Set the selected language into a cookie."""


class ISchoolToolInitializationUtility(Interface):

    def initializeApplication(app):
        """Perform school specific initialization. """


class IPluginAction(Interface):

    after = Attribute(
        """A list of action adapter names.
        This action must be executed after them.""")

    before = Attribute(
        """A list of action adapter names.
        This action must be executed before them.""")

    def __call__():
        """Perform plugin specific set up."""


class ICatalogStartUp(IPluginAction):
    """Set up SchoolTool catalogs."""


class IPluginInit(IPluginAction):
    """Perform plugin initialization when setting up the SchoolTool
    application."""


class IPluginStartUp(IPluginAction):
    """Execute plugin specific code on application startup."""


class ISchoolToolAuthenticationPlugin(ISchoolToolAuthentication):
    """A plugin for local schooltool authentication utility. """


class IVersionedCatalog(IContained):
    """Versioned catalog entry."""

    version = zope.schema.TextLine(
        title=u"Version",
        description=u"Current version of the catalog.")

    expired = zope.schema.Bool(
        title=u"Expired",
        description=u"Expired catalogs should be removed.")

    catalog = Attribute("""The catalog object.""")


class ICatalogs(IContainer):
    """Container of versioned catalogs."""

    contains(IVersionedCatalog)




#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2011 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Report interfaces

"""

import zope.schema
import zope.file.interfaces
from zope.interface import Attribute, Interface
from zope.location.interfaces import ILocation
from zope.publisher.interfaces.browser import IBrowserRequest
from zope.publisher.interfaces.browser import IBrowserPage
from zope.viewlet.interfaces import IViewletManager

from schooltool.task.interfaces import IRemoteTask
from schooltool.task.interfaces import IProgressMessage
from schooltool.task.interfaces import IMessage
from schooltool.skin.flourish.interfaces import IFlourishLayer

from schooltool.common import SchoolToolMessage as _


class IReportLinkViewletManager(IViewletManager):
   """The manager for report links."""


class IReportLinkViewlet(Interface):
   """A report link viewlet"""


class IRegisteredReportsUtility(Interface):

    reports_by_group = zope.schema.Dict(
        title=u"Reports by group",
        description=u"Maps report group names to lists of report descriptions")


class IReportLinksURL(Interface):

    def actualContext():
        """Returns the actual object the report links are for."""

    def __unicode__():
        """Returns the URL as a unicode string."""

    def __str__():
        """Returns an ASCII string with all unicode characters url quoted."""

    def __repr__():
        """Get a string representation """

    def __call__():
        """Returns an ASCII string with all unicode characters url quoted."""


class IReportTask(IRemoteTask):

   factory = Attribute('Report factory class or callable')
   factory_name = zope.schema.TextLine(
      title=u'Report callable signature',
      required=False)
   view_name = zope.schema.TextLine(
      title=u'Report view name',
      required=False)
   context = Attribute('Context object')

   default_filename = zope.schema.TextLine(
      title=u'Default report filename',
      required=False)
   default_mimetype = zope.schema.TextLine(
      title=u'Default report mime type',
      required=False)


class IReportFile(zope.file.interfaces.IFile):
   pass


class IReportDetails(Interface):

    report = zope.schema.Object(
       title=_('Report'),
       schema=IReportFile,
       required=False)

    filename = zope.schema.TextLine(
       title=_('Filename'),
       required=False)

    requested_on = zope.schema.Datetime(
       title=_("Requested on"),
       required=False)


class IReportProgressMessage(IProgressMessage, IReportDetails):
   pass


class IReportMessage(IMessage, IReportDetails):
   pass


class IRemoteReportLayer(IFlourishLayer, IBrowserRequest):
   pass


class IArchivePage(IBrowserPage, ILocation):

    def update(self):
        pass

    def render(stream, *args, **kw):
        """Render the archive to stream."""



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
CSV import view for BasicPerson.
"""
from zope.component import getUtility
from zope.container.interfaces import INameChooser
from zope.traversing.browser.absoluteurl import absoluteURL

from schooltool.app.browser.csvimport import BaseCSVImporter
from schooltool.app.browser.csvimport import BaseCSVImportView
from schooltool.app.browser.csvimport import FlourishBaseCSVImportView
from schooltool.person.interfaces import IPersonFactory

from schooltool.common import SchoolToolMessage as _


class BasicPersonCSVImporter(BaseCSVImporter):
    """A Person CSV importer."""

    def createAndAdd(self, data, dry_run=True):
        """Create Person object and add to container.

        We are requiring that we have a username and fullname (title) set.  If
        any duplicates are found then an error is reported and the duplicate
        entries are reported back to the user.
        """
        if len(data) < 3:
            self.errors.fields.append(_("""Insufficient data provided."""))
            return

        if not data[0]:
            self.errors.fields.append(_('username may not be empty'))
            return

        if not data[1]:
            self.errors.fields.append(_('first name may not be empty'))
            return

        if not data[2]:
            self.errors.fields.append(_('last name may not be empty'))
            return

        username = data[0]
        first_name = data[1]
        last_name = data[2]
        if len(data) > 3:
            password = data[3]
        else:
            password = None

        if username in self.container:
            error_msg = _("Duplicate username: ${username}",
                          mapping={'username' : ', '.join(data)})
            self.errors.fields.append(error_msg)
            return

        try:
            INameChooser(self.container).checkName(username, None)
        except ValueError:
            error_msg = _("Names cannot begin with '+' or '@' or contain '/'")
            self.errors.fields.append(error_msg)
            return

        # XXX: this has to be fixed
        # XXX: SchoolTool should handle UTF-8
        try:
            username.encode('ascii')
        except UnicodeEncodeError:
            error_msg = _("Usernames cannot contain non-ascii characters")
            self.errors.fields.append(error_msg)
            return

        obj = self.personFactory(username, first_name, last_name)

        if password:
            obj.setPassword(password)

        if not dry_run:
            self.container[data[0]] = obj

    def personFactory(self, username, first_name, last_name):
        factory = getUtility(IPersonFactory)
        person = factory(username=username,
                         first_name=first_name,
                         last_name=last_name)
        return person


class BasicPersonCSVImportView(BaseCSVImportView):
    """View for Person CSV importer."""

    importer_class = BasicPersonCSVImporter


class FlourishBasicPersonCSVImportView(FlourishBaseCSVImportView):

    importer_class = BasicPersonCSVImporter

    def nextURL(self):
        return absoluteURL(self.context, self.request)



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Unit tests for schooltool.note
"""
import unittest
import doctest

from zope.interface import implements
from zope.publisher.interfaces.browser import IBrowserRequest
from zope.publisher.browser import TestRequest
from zope.component import provideAdapter

from zope.app.testing import setup

from schooltool.app.browser.testing import setUp, tearDown


def doctest_getNotes():
    r"""Test for schooltool.note.note.getNotes.

    We need to set up Zope 3 annotations

        >>> setup.placelessSetUp()
        >>> setup.setUpAnnotations()

    We need to have an annotatable object

        >>> from zope.annotation.interfaces import IAttributeAnnotatable
        >>> class SomeAnnotatable(object):
        ...     implements(IAttributeAnnotatable)

        >>> obj = SomeAnnotatable()

    Now we can check that a new Notes object is created automatically

        >>> from schooltool.note.note import getNotes
        >>> notes = getNotes(obj)

        >>> from schooltool.note.interfaces import INotes
        >>> from zope.interface.verify import verifyObject
        >>> verifyObject(INotes, notes)
        True

    If you do it more than once, you will get the same Notes object

        >>> notes is getNotes(obj)
        True

    """

def doctest_browser_NoteAddView():
    r"""Test for NoteAddView

    We need some setup to make traversal work in a unit test.

        >>> class FakeURL:
        ...     def __init__(self, context, request): pass
        ...     def __call__(self):
        ...         return "http://127.0.0.1/frogpond/persons/milton"
        >>> from schooltool.note.interfaces import IHaveNotes, INotes
        >>> from zope.traversing.browser.interfaces import IAbsoluteURL
        >>> provideAdapter(FakeURL,
        ...                provides=IAbsoluteURL,
        ...                adapts=(IHaveNotes, IBrowserRequest))

    Let's create an owner for our note.  We need an id for our person to test
    the AddView later on:

        >>> from zope.annotation.interfaces import IAttributeAnnotatable
        >>> class Owner(object):
        ...     implements(IAttributeAnnotatable, IHaveNotes)
        ...     id = None

        >>> owner = Owner()
        >>> owner.id = 'someone'

    Now let's create a NoteAddView for the owner

        >>> from schooltool.note.browser import NoteAddView
        >>> view = NoteAddView(owner, TestRequest())
        >>> view.update()

    We need to set up an adapter for notes

        >>> setup.setUpAnnotations()
        >>> from schooltool.note.note import getNotes
        >>> provideAdapter(getNotes, adapts=[IHaveNotes], provides=INotes)

    Let's try to add a note:

        >>> request = TestRequest(form={'field.title': u'Red Stapler',
        ...                             'field.body': u"He won't give it back",
        ...                             'field.privacy': u"public",
        ...                             'UPDATE_SUBMIT': 'Add'})
        >>> request.setPrincipal(owner)
        >>> view = NoteAddView(owner, request)
        >>> view.update()
        ''
        >>> print view.errors
        ()
        >>> print view.error
        None
        >>> notes = INotes(owner)
        >>> print [note.title for note in notes.__iter__()]
        [u'Red Stapler']

    """


def test_suite():
    return unittest.TestSuite([
        doctest.DocTestSuite(setUp=setUp, tearDown=tearDown,
                             optionflags=doctest.ELLIPSIS|
                             doctest.REPORT_NDIFF),
        doctest.DocTestSuite('schooltool.note.note'),
           ])

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Calendar overlay views for the SchoolTool application.
"""

import urllib

from zope.browserpage import ViewPageTemplateFile
from zope.cachedescriptors.property import Lazy
from zope.interface import implements, Interface
from zope.publisher.browser import BrowserView
from zope.traversing.api import getPath
from zope.traversing.browser.absoluteurl import absoluteURL
from zope.location.interfaces import ILocation
from zope.security.proxy import removeSecurityProxy
from zope.security.checker import canAccess
from zope.viewlet.viewlet import ViewletBase
from zc.table.column import GetterColumn

from schooltool.common import SchoolToolMessage as _
from schooltool.app.interfaces import ISchoolToolCalendar
from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.app.browser.app import FlourishRelationshipViewBase
from schooltool.app.browser.app import EditRelationships
from schooltool.app.browser.app import RelationshipAddTableMixin
from schooltool.app.browser.app import RelationshipRemoveTableMixin
from schooltool.resource.interfaces import IResourceTypeInformation
from schooltool.person.interfaces import IPerson
from schooltool import table
from schooltool.table.table import SchoolToolTableFormatter
from schooltool.table.table import FilterWidget


class CalendarOverlayBase(object):
    """View for the calendar overlay portlet.

    This view can be used with any context, but it gets rendered to an empty
    string unless context is the calendar of the authenticated user.

    Note that this view contains a self-posting form and handles submits that
    contain 'OVERLAY_APPLY' or 'OVERLAY_MORE' in the request.
    """

    def show_overlay(self):
        """Check whether the calendar overlay portlet needs to be rendered.

            >>> from zope.app.testing import setup
            >>> setup.placelessSetUp()
            >>> setup.setUpAnnotations()

            >>> from schooltool.testing import setup as sbsetup
            >>> sbsetup.setUpCalendaring()

        The portlet is only shown when an authenticated user is looking
        at his/her calendar.

        Anonymous user:

            >>> from zope.publisher.browser import TestRequest
            >>> from schooltool.person.person import Person
            >>> request = TestRequest()
            >>> person = Person()
            >>> context = ISchoolToolCalendar(person)
            >>> view = CalendarOverlayView(context, request, None, None)
            >>> view.show_overlay()
            False

        Person that we're looking at

            >>> from schooltool.app.security import Principal
            >>> request.setPrincipal(Principal('id', 'title', person))
            >>> view.show_overlay()
            True

        A different person:

            >>> request.setPrincipal(Principal('id', 'title', Person()))
            >>> view.show_overlay()
            False

       Cleanup:

            >>> setup.placelessTearDown()

        """
        if not ILocation.providedBy(self.context):
            return False
        logged_in = removeSecurityProxy(IPerson(self.request.principal, None))
        calendar_owner = removeSecurityProxy(self.context.__parent__)
        return logged_in is calendar_owner

    def items(self):
        """Return items to be shown in the calendar overlay.

        Does not include "my calendar".

        Each item is a dict with the following keys:

            'title' - title of the calendar

            'calendar' - the calendar object

            'color1', 'color2' - colors assigned to this calendar

            'id' - identifier for form controls

            'checked' - was this item checked for display (either "checked" or
            None)?

        """
        person = IPerson(self.request.principal)
        items = [((item.calendar.title, getPath(item.calendar.__parent__)),
                  {'title': item.calendar.title,
                   'id': getPath(item.calendar.__parent__),
                   'calendar': item.calendar,
                   'checked': item.show and "checked" or '',
                   'color1': item.color1,
                   'color2': item.color2})
                 for item in person.overlaid_calendars
                 if canAccess(item.calendar, '__iter__')]
        items.sort()
        return [i[-1] for i in items]

    def update(self):
        """Process form submission."""
        if 'OVERLAY_MORE' in self.request:
            person = IPerson(self.request.principal)
            url = absoluteURL(person, self.request)
            url += '/calendar_selection.html'
            url += '?nexturl=%s' % urllib.quote(str(self.request.URL))
            self.request.response.redirect(url)
        if 'OVERLAY_APPLY' in self.request:
            person = IPerson(self.request.principal)
            selected = set(self.request.get('overlay', []))
            for item in person.overlaid_calendars:
                item.show = getPath(item.calendar.__parent__) in selected
            url = str(self.request.URL)
            self.request.response.redirect(url)


class CalendarOverlayView(CalendarOverlayBase, ViewletBase):
    pass


class CalendarSelectionView(BrowserView):
    """A view for calendar selection.

    This view can be used with any context, but always operates on the
    currently authenticated user's list of overlaid calendars.
    """

    error = None
    message = None

    def getCalendars(self, container):
        """List all calendars from a given container."""
        user = removeSecurityProxy(IPerson(self.request.principal, None))
        if user is None:
            return []
        app = ISchoolToolApplication(None)

        result = []
        for obj in app[container].values():
            calendar = ISchoolToolCalendar(obj)
            if obj is not user and canAccess(calendar, '__iter__'):
                result.append(
                    {'id': obj.__name__,
                     'title': obj.title,
                     'selected': calendar in user.overlaid_calendars,
                     'calendar': calendar})
        return sorted(result, key=lambda item: (item['title'], item['id']))

    def getApplicationCalendar(self):
        """Return the application calendar.

        Returns None if the user lacks sufficient permissions.
        """
        user = IPerson(self.request.principal, None)
        if user:
            app = ISchoolToolApplication(None)
            calendar = ISchoolToolCalendar(app)
            if canAccess(calendar, '__iter__'):
                return {'title': app.title,
                        'selected': calendar in user.overlaid_calendars,
                        'calendar': calendar}
        return {}

    application = property(getApplicationCalendar)
    resources = property(lambda self: self.getCalendars('resources'))

    def update(self):
        """Process forms."""
        if 'CANCEL' in self.request:
            nexturl = self.request.form.get('nexturl')
            if nexturl:
                self.request.response.redirect(nexturl)
            return
        user = IPerson(self.request.principal, None)
        if user is None:
            return
        if 'UPDATE_SUBMIT' in self.request:
            self._updateSelection(user)
            self.message = _('Saved changes.')
            nexturl = self.request.form.get('nexturl')
            if nexturl:
                self.request.response.redirect(nexturl)

    def _updateSelection(self, user):
        """Apply calendar selection changes  for `user`."""
        for container in ['resources']:
            selected = set(self.request.form.get(container, []))
            for item in self.getCalendars(container):
                if item['id'] in selected and not item['selected']:
                    user.overlaid_calendars.add(item['calendar'])
                elif item['id'] not in selected and item['selected']:
                    user.overlaid_calendars.remove(item['calendar'])
        appcal = self.getApplicationCalendar().get('calendar')
        if appcal is not None:
            if ('application' in self.request and
                    appcal not in user.overlaid_calendars):
                user.overlaid_calendars.add(appcal)
            elif ('application' not in self.request and
                    appcal in user.overlaid_calendars):
                user.overlaid_calendars.remove(appcal)


class IOverlayCalendarsContainer(Interface):
    pass


class OverlayCalendarsInfo(object):
    __name__ = None
    def __init__(self, calendar, title, cal_type):
        self.calendar = calendar
        self.title = title
        self.cal_type = cal_type

class OverlayCalendarsContainer(dict):
    implements(IOverlayCalendarsContainer)

    def __init__(self):
        self.by_calendar = {}

    def __setitem__(self, key, info):
        dict.__setitem__(self, key, info)
        info.__name__ = key
        self.by_calendar[info.calendar] = info


class OverlayCalendarsFormatter(SchoolToolTableFormatter):

    def columns(self):
        title = GetterColumn(
            name='title',
            title=_(u"Title"),
            getter=lambda i, f: i.title,
            subsort=True)
        type = GetterColumn(
            name='type',
            title=_(u"Type"),
            getter=lambda i, f: i.cal_type,
            subsort=True)
        return [title, type]


class OverlayCalendarsFilterWidget(FilterWidget):
    template = ViewPageTemplateFile('templates/f_overlay_filter_widget.pt')


class OverlayCalendarTable(table.ajax.Table):

    def columns(self):
        title = GetterColumn(
            name='title',
            title=_(u"Title"),
            getter=lambda i, f: i.title,
            subsort=True)
        type = GetterColumn(
            name='type',
            title=_(u"Type"),
            getter=lambda i, f: i.cal_type,
            subsort=True)
        return [title, type]


class CalendarAddRelationshipTable(RelationshipAddTableMixin,
                                   OverlayCalendarTable):
    pass


class CalendarRemoveRelationshipTable(RelationshipRemoveTableMixin,
                                      OverlayCalendarTable):
    pass


class FlourishCalendarSelectionView(EditRelationships):

    current_title = _('Selected calendars')
    available_title = _('Available calendars')

    @Lazy
    def user(self):
        return IPerson(self.request.principal, None)

    def getCollection(self):
        if self.user is None:
            return []
        return self.user.overlaid_calendars

    def getResourceCalendars(self):
        if self.user is None:
            return []
        app = ISchoolToolApplication(None)
        result = []
        for obj in app['resources'].values():
            calendar = ISchoolToolCalendar(obj)
            if canAccess(calendar, '__iter__'):
                result.append(calendar)
        return result

    def getApplicationCalendar(self):
        if self.user is None:
            return None
        app = ISchoolToolApplication(None)
        calendar = ISchoolToolCalendar(app)
        if not canAccess(calendar, '__iter__'):
            return None
        return calendar

    @Lazy
    def overlay_container(self):
        container = OverlayCalendarsContainer()
        cal = self.getApplicationCalendar()
        if cal is not None:
            key = 'app.school'
            container[key] = OverlayCalendarsInfo(
                removeSecurityProxy(cal),
                cal.__parent__.title, _('School Calendar'))
        for cal in self.getResourceCalendars():
            resource = cal.__parent__
            key = 'resource.%s' % resource.__name__
            res_info = IResourceTypeInformation(resource)
            container[key] = OverlayCalendarsInfo(
                removeSecurityProxy(cal),
                resource.title, res_info.title)
        return container

    def getAvailableItemsContainer(self):
        return self.overlay_container

    def add(self, item):
        super(FlourishCalendarSelectionView, self).add(item.calendar)

    def remove(self, item):
        super(FlourishCalendarSelectionView, self).remove(item.calendar)

    def getAvailableItems(self):
        available = self.getAvailableItemsContainer()
        selected = set([removeSecurityProxy(p.calendar)
                        for p in self.getCollection()])
        return [info
                for cal, info in available.by_calendar.items()
                if cal not in selected]

    def getSelectedItems(self):
        available = self.getAvailableItemsContainer()
        selected = set([removeSecurityProxy(p.calendar)
                        for p in self.getCollection()])
        return [available.by_calendar[calendar]
                for calendar in selected
                if calendar in available.by_calendar]

    def nextURL(self):
        url = self.request.get('nexturl')
        if url is None:
            cal = ISchoolToolCalendar(self.context)
            url = absoluteURL(cal, self.request)
        return url



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Calendar overlays.

This module defines relationships used to store calendar subscriptions.

    >>> from zope.app.testing import setup
    >>> setup.placelessSetUp()
    >>> setup.setUpAnnotations()

    >>> from schooltool.testing import setup as sbsetup
    >>> sbsetup.setUpCalendaring()

    >>> from schooltool.relationship.tests import setUpRelationships
    >>> setUpRelationships()

We will need some sample persons and groups for the demonstration

    >>> from schooltool.group.group import Group
    >>> from schooltool.person.person import Person
    >>> john = Person(title="John")
    >>> smith = Person(title="Smith")
    >>> developers = Group(title="Developers")
    >>> admins = Group(title="Admins")

Let's say John wants to see the calendars of Smith and the Developers group
overlaid on his own calendar

    >>> john.overlaid_calendars.add(ISchoolToolCalendar(smith))
    <...CalendarOverlayInfo object at ...>
    >>> john.overlaid_calendars.add(ISchoolToolCalendar(developers))
    <...CalendarOverlayInfo object at ...>

He also wants the Admins group calendar to be displayed in the overlaid
calendars portlet, but hidden by default, and perhaps set the colour to green:

    >>> info = john.overlaid_calendars.add(ISchoolToolCalendar(admins),
    ...                                    show=False)
    >>> info.color1 = 'green'

Iterating over `overlaid_calendars` returns ICalendarOverlayInfo objects

    >>> for item in john.overlaid_calendars:
    ...     checked = item.show and "+" or " "
    ...     title = item.calendar.__parent__.title
    ...     print "[%s] %-10s (%s, %s)" % (checked, title,
    ...                                    item.color1, item.color2)
    [+] Smith      (#e0b6af, #c1665a)
    [+] Developers (#eed680, #d1940c)
    [ ] Admins     (green, #83a67f)

However, 'in' checks for the presence of a calendar

    >>> ISchoolToolCalendar(smith) in john.overlaid_calendars
    True
    >>> ISchoolToolCalendar(Person(title="Newcomer")) in john.overlaid_calendars
    False

Clean up

    >>> setup.placelessTearDown()

"""

import sys

from persistent import Persistent
from zope.proxy import getProxiedObject
from zope.interface import Interface, implements
from zope.schema import Object, TextLine, Bool
from zope.lifecycleevent.interfaces import IObjectRemovedEvent

from schooltool.relationship import URIObject
from schooltool.relationship.interfaces import IRelationshipLinks
from schooltool.relationship.relationship import BoundRelationshipProperty
from schooltool.app.interfaces import ISchoolToolCalendar
from schooltool.app.interfaces import IHaveCalendar
from schooltool.relationship.relationship import unrelateAll


URICalendarSubscription = URIObject(
                "http://schooltool.org/ns/calendar_subscription",
                "Calendar subscription",
                "The calendar subscription relationship.")

URICalendarProvider = URIObject(
                "http://schooltool.org/ns/calendar_subscription/provider",
                "Calendar provider",
                "A role of an object providing a calendar.")

URICalendarSubscriber = URIObject(
                "http://schooltool.org/ns/calendar_subscription/subscriber",
                "Calendar subscriber",
                "A role of an object that subscribes to a calendar.")


DEFAULT_COLORS = (
        ('#e0b6af', '#c1665a'), # Red Highlight, Red Medium
        ('#eed680', '#d1940c'), # Accent Yellow, Accent Yellow Dark
        ('#c5d2c8', '#83a67f'), # Green Highlight, Green Medium
        ('#efe0cd', '#e0c39e'), # Face Skin Highlight, Face Skin Medium
        ('#ada7c8', '#887fa3'), # Purple Highlight, Purple Medium
        ('#eae8e3', '#bab5ab'), # Basic 3D Highlight, Basic 3D Medium
        ('#e0c39e', '#b39169'), # Face Skin Medium, Face Skin Dark
        ('#c1665a', '#884631'), # Red Medium, Red Dark
        ('#b39169', '#826647'), # Face Skin Dark, Face Skin Shadow
        ('#83a67f', '#5d7555'), # Green Medium, Green Dark
    )


class ICalendarOverlayInfo(Interface):
    """Information about an overlaid calendar."""

    calendar = Object(
            title=u"Calendar",
            schema=ISchoolToolCalendar,
            description=u"""
            Calendar.
            """)

    color1 = TextLine(
            title=u"Color 1",
            description=u"""
            Color for this calendar.

            This is a string that is acceptable as a CSS color, e.g. '#ccffee'.
            """)

    color2 = TextLine(
            title=u"Color 2",
            description=u"""
            Color for this calendar.

            This is a string that is acceptable as a CSS color, e.g. '#ccffee'.
            """)

    show = Bool(
            title=u"Show",
            description=u"""
            An option that controls whether events from this calendar are shown
            in the calendar views (show=True), or if they are only listed in
            the portlet (show=False).
            """)


class IOverlaidCalendarsProperty(Interface):
    """A property for maintaining a list of overlaid calendars."""

    def __nonzero__():
        """Are there any overlaid calendars?"""

    def __len__():
        """Return the number of overlaid calendars."""

    def __contains__(calendar):
        """Check whether `calendar` is in the list."""

    def __iter__():
        """Iterate over all overlaid calendars.

        Returns ICalendarOverlayInfo objects.  Iteration order is unspecified.
        """

    def add(calendar, show=True, color1=None, color2=None):
        """Add `calendar` to the list.

        If `color1` or `color2` is not specified, a pair of colours are chosen
        from a list of standard colors.  The color chooser tries to minimize
        color conflicts with other overlaid calendars.
        """

    def remove(calendar):
        """Remove `calendar` from the list."""


class OverlaidCalendarsProperty(object):
    """Property for `overlaid_calendars`.

    Stores the list of overlaid calendars in relationships.

    Example:

        >>> class SomeClass(object): # must be a new-style class
        ...     calendars = OverlaidCalendarsProperty()

        >>> from zope.interface.verify import verifyObject
        >>> someinstance = SomeClass()
        >>> verifyObject(IOverlaidCalendarsProperty, someinstance.calendars)
        True

    """

    def __get__(self, instance, owner):
        if instance is None:
            return self
        else:
            return BoundOverlaidCalendarsProperty(instance)


class BoundOverlaidCalendarsProperty(BoundRelationshipProperty):
    """Bound property for `overlaid_calendars`

    Stores the list of overlaid calendars in relationships.

        >>> from schooltool.relationship.tests import setUp, tearDown
        >>> from schooltool.relationship.tests import SomeObject
        >>> from schooltool.relationship import getRelatedObjects
        >>> setUp()

    Given a relatable object, and a relatable calendar

        >>> a = SomeObject('a')
        >>> cal = SomeObject('cal')

    we can create a BoundOverlaidCalendarsProperty

        >>> overlaid_calendars = BoundOverlaidCalendarsProperty(a)

    The `add` method establishes a URICalendarSubscriber relationship

        >>> overlaid_calendars.add(cal)
        <...CalendarOverlayInfo object at ...>
        >>> getRelatedObjects(a, URICalendarProvider)
        [cal]
        >>> getRelatedObjects(cal, URICalendarSubscriber)
        [a]

    `__nonzero__` and `__len__` do the obvious things

        >>> bool(overlaid_calendars)
        True
        >>> len(overlaid_calendars)
        1

    The `remove` method breaks it

        >>> overlaid_calendars.remove(cal)
        >>> len(overlaid_calendars)
        0
        >>> bool(overlaid_calendars)
        False
        >>> getRelatedObjects(a, URICalendarProvider)
        []
        >>> getRelatedObjects(cal, URICalendarSubscriber)
        []

    You can specify extra arguments for `add`

        >>> overlaid_calendars.add(cal, show=False,
        ...                        color1="red", color2="green")
        <...CalendarOverlayInfo object at ...>

    You can extract these when iterating

        >>> for item in overlaid_calendars:
        ...     print item.calendar, item.show, item.color1, item.color2
        cal False red green

    We're done.

        >>> tearDown()

    """

    implements(IOverlaidCalendarsProperty)

    def __init__(self, this):
        BoundRelationshipProperty.__init__(self, this, URICalendarSubscription,
                                           URICalendarSubscriber,
                                           URICalendarProvider)
        # for the local grants to work
        self.__parent__ = this

    def add(self, calendar, show=True, color1=None, color2=None):
        if not color1 or not color2:
            used_colors = [(item.color1, item.color2) for item in self]
            color1, color2 = choose_color(DEFAULT_COLORS, used_colors)
        info = CalendarOverlayInfo(calendar, show, color1, color2)
        info.__parent__ = self.this
        BoundRelationshipProperty.add(self, calendar, info)
        return info

    def __contains__(self, calendar):
        for item in self:
            if item.calendar is calendar:
                return True
        return False

    def __iter__(self):
        for link in IRelationshipLinks(self.this):
            if link.role == self.other_role and link.rel_type == self.rel_type:
                yield link.extra_info


class CalendarOverlayInfo(Persistent):
    """Information about an overlaid calendar.

        >>> from zope.interface.verify import verifyObject
        >>> calendar = object()
        >>> item = CalendarOverlayInfo(calendar, True, 'red', 'yellow')
        >>> verifyObject(ICalendarOverlayInfo, item)
        True

    The calendar attribute must be read-only, because a CalendarOverlayInfo is
    stored as an attribute on a specific relationship with a specific calendar
    object.

        >>> item.calendar = object()
        Traceback (most recent call last):
          ...
        AttributeError: can't set attribute

    `show`, `color1` and `color2` attributes are changeable

        >>> item.show = True
        >>> item.color1 = 'blue'
        >>> item.color2 = 'black'

    """

    implements(ICalendarOverlayInfo)

    calendar = property(lambda self: self._calendar)

    def __init__(self, calendar, show, color1, color2):
        self._calendar = calendar
        self.show = show
        self.color1 = color1
        self.color2 = color2


def choose_color(colors, used_colors):
    """Choose a color, avoiding colors that have been used already.

        >>> colors = ('red', 'green', 'blue')

    choose_color picks the first unused color.

        >>> choose_color(colors, [])
        'red'
        >>> choose_color(colors, ['red'])
        'green'
        >>> choose_color(colors, ['green', 'red'])
        'blue'

    If all colors have been used, choose_color picks the one that
    has been used the least number of times, and if there are several
    such colors, picks the first of them.

        >>> choose_color(colors, ['green', 'red', 'blue'])
        'red'
        >>> choose_color(colors, ['green', 'red', 'blue', 'red'])
        'green'

    You can also use choose_color for color pairs

        >>> pairs = [('red', 'green'), ('red', 'yellow'), ('blue', 'green')]
        >>> choose_color(pairs, [('red', 'green')])
        ('red', 'yellow')

    """
    if not colors:
        raise ValueError("no colors to choose from")
    used_count = {}
    for c in used_colors:
        used_count[c] = used_count.get(c, 0) + 1
    min_count = sys.maxint
    for c in colors:
        count = used_count.get(c, 0)
        if count == 0:
            return c
        if count < min_count:
            best_color = c
            min_count = count
    return best_color


def unrelateCalendarOnDeletion(event):
    """When you delete an object, relationships of it's calendar should be removed

        >>> from schooltool.relationship.tests import setUp, tearDown
        >>> from schooltool.testing.setup import setUpCalendaring

        >>> setUp()
        >>> setUpCalendaring()

        >>> import zope.event
        >>> old_subscribers = zope.event.subscribers[:]
        >>> from schooltool.app.overlay import unrelateCalendarOnDeletion
        >>> zope.event.subscribers.append(unrelateCalendarOnDeletion)


    We will need some object that implements IHaveCalendar for that:

        >>> from zope.container.btree import BTreeContainer
        >>> container = BTreeContainer()
        >>> from schooltool.person.person import Person
        >>> container = BTreeContainer()
        >>> container['jonas'] = jonas = Person(username="Jonas")
        >>> container['petras'] = petras =  Person(username="Petras")

    Let's add calendar of Petras to the list of overlaid calendars:

        >>> jonas.overlaid_calendars.add(ISchoolToolCalendar(petras))
        <...CalendarOverlayInfo object at ...>
        >>> list(jonas.overlaid_calendars)
        [<schooltool.app.overlay.CalendarOverlayInfo object at ...>]

    If we delete Petras - Jonas should have no calendars in his overlay list:

        >>> del container['petras']
        >>> list(jonas.overlaid_calendars)
        []

    Restore old subscribers:

        >>> zope.event.subscribers[:] = old_subscribers
        >>> tearDown()

    """
    if not IObjectRemovedEvent.providedBy(event):
        return
    # event.object may be a ContainedProxy
    obj = getProxiedObject(event.object)
    if not IHaveCalendar.providedBy(obj):
        return
    calendar = ISchoolToolCalendar(obj)
    linkset = IRelationshipLinks(calendar, None)
    if linkset is not None:
        unrelateAll(calendar)



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Timetabling Schema views.
"""
import urllib

import zope.schema
import zope.lifecycleevent
from zope.proxy import sameProxiedObjects
from zope.component import getMultiAdapter
from zope.component import adapts
from zope.component import getUtility
from zope.container.interfaces import INameChooser
from zope.interface import Interface, implements
from zope.intid.interfaces import IIntIds
from zope.i18n import translate
from zope.browserpage.viewpagetemplatefile import ViewPageTemplateFile
from zope.publisher.browser import BrowserView
from zope.publisher.interfaces.browser import IBrowserRequest
from zope.publisher.interfaces.browser import IBrowserView
from zope.security.proxy import removeSecurityProxy
from zope.traversing.browser.absoluteurl import absoluteURL
from z3c.form import form, field, button, widget, validator
from z3c.form.util import getSpecification
from z3c.form.browser.checkbox import SingleCheckBoxFieldWidget

import schooltool.skin.flourish.page
from schooltool.skin import flourish
from schooltool.skin.flourish.content import ContentProvider
from schooltool.common import format_time_range
from schooltool.table.table import simple_form_key
from schooltool.timetable.interfaces import ITimetable, ITimetableContainer
from schooltool.timetable.interfaces import ISelectedPeriodsSchedule
from schooltool.timetable.interfaces import IHaveSchedule, IScheduleContainer
from schooltool.timetable.browser.app import getActivityVocabulary
from schooltool.timetable.browser.schedule import FlourishConfirmDeleteView
from schooltool.timetable.timetable import SelectedPeriodsSchedule
from schooltool.term.interfaces import ITerm

from schooltool.common import SchoolToolMessage as _


class IRenderDayTableCells(IBrowserView):
    def renderCells(schedule, day, item):
        """Return contents for two <td> cells: (title_html, value_html)
           Or return None if don't want to render the cell.
        """


class DayTemplatesTable(ContentProvider):

    @property
    def days(self):
        return self.context.templates.values()

    def makeTable(self):
        days = self.days
        table = {'header': [day.title for day in days]}

        def to_dict(item):
            return item and {'title': item[0], 'value': item[1]} or {}

        cols = []
        for day in days:
            cells = [self.view.renderCells(self.context, day, item)
                     for item in day.values()]
            cols.append(map(to_dict, filter(None, cells)))

        max_rows = max([len(cells) for cells in cols])
        cols = [cells + [{}]*(max_rows-len(cells)) for cells in cols]

        if len(cols) == 1:
            table['rows'] = [(col,) for col in cols[0]]
        else:
            table['rows'] = map(None, *cols)
        ncols = len(cols) or 1
        table['col_width'] ='%d%%' % (100 / ncols);
        table['th_width'] = '%d%%' % (10 / ncols);
        table['td_width'] = '%d%%' % (90 / ncols);
        return table


class TimetableView(BrowserView):
    implements(IRenderDayTableCells)
    adapts(ITimetable, IBrowserRequest)

    template = ViewPageTemplateFile("templates/timetable.pt")

    activity_vocabulary = None

    def __init__(self, *args, **kw):
        BrowserView.__init__(self, *args, **kw)
        self.activity_vocabulary = getActivityVocabulary(self.context)

    def activityTitle(self, activity_type):
        if activity_type in self.activity_vocabulary:
            term = self.activity_vocabulary.getTerm(activity_type)
            return term.title
        return None

    def renderCells(self, schedule, day, item):
        if sameProxiedObjects(schedule, self.context.periods):
            period = item
            return (period.title,
                    self.activityTitle(period.activity_type) or '')
        if sameProxiedObjects(schedule, self.context.time_slots):
            slot = item
            return (format_time_range(slot.tstart, slot.duration),
                    self.activityTitle(slot.activity_type) or '')
        return None

    def __call__(self):
        return self.template()


class FlourishTimetableView(flourish.page.WideContainerPage, TimetableView):
    pass


#class SimpleTimetableSchemaAdd(BrowserView):
#    """A simple timetable schema definition view"""
#
#    _nrperiods = 9
#
#    day_ids = (_("Monday"),
#               _("Tuesday"),
#               _("Wednesday"),
#               _("Thursday"),
#               _("Friday"),
#               )
#
#    error = None
#
#    template = ViewPageTemplateFile('templates/simpletts.pt')
#
#    def __init__(self, content, request):
#        BrowserView.__init__(self, content, request)
#        self._schema = {}
#        self._schema['title'] = TextLine(__name__='title', title=_(u"Title"))
#        for nr in range(1, self._nrperiods + 1):
#            pname = 'period_name_%s' % nr
#            pstart = 'period_start_%s' % nr
#            pfinish = 'period_finish_%s' % nr
#            self._schema[pname] = TextLine(__name__=pname,
#                                           title=u"Period title",
#                                           required=False)
#            self._schema[pstart] = TextLine(__name__=pstart,
#                                            title=u"Period start time",
#                                            required=False)
#            self._schema[pfinish] = TextLine(__name__=pfinish,
#                                             title=u"Period finish time",
#                                             required=False)
#        setUpWidgets(self, self._schema, IInputWidget,
#                     initial={'title': 'default'})
#
#    def _setError(self, name, error=RequiredMissing()):
#        """Set an error on a widget."""
#        # XXX Touching widget._error is bad, see
#        #     http://dev.zope.org/Zope3/AccessToWidgetErrors
#        # The call to setRenderedValue is necessary because
#        # otherwise _getFormValue will call getInputValue and
#        # overwrite _error while rendering.
#        widget = getattr(self, name + '_widget')
#        widget.setRenderedValue(widget._getFormValue())
#        if not IWidgetInputError.providedBy(error):
#            error = WidgetInputError(name, widget.label, error)
#        widget._error = error
#
#    def getPeriods(self):
#        try:
#            data = getWidgetsData(self, self._schema)
#        except WidgetsError:
#            return []
#
#        result = []
#        for nr in range(1, self._nrperiods + 1):
#            pname = 'period_name_%s' % nr
#            pstart = 'period_start_%s' % nr
#            pfinish = 'period_finish_%s' % nr
#            if data.get(pstart) or data.get(pfinish):
#                try:
#                    start, duration = parse_time_range(
#                        "%s-%s" % (data[pstart], data[pfinish]))
#                except ValueError:
#                    self.error = _('Please use HH:MM format for period '
#                                   'start and end times')
#                    continue
#                name = data[pname]
#                if not name:
#                    name = data[pstart]
#                result.append((name, start, duration))
#        return result
#
#    def createSchema(self, periods):
#        daytemplate = SchooldayTemplate()
#        for title, start, duration in periods:
#            daytemplate.add(SchooldaySlot(start, duration))
#
#        factory = getUtility(ITimetableModelFactory, 'WeeklyTimetableModel')
#        model = factory(self.day_ids, {None: daytemplate})
#        app = ISchoolToolApplication(None)
#        tzname = IApplicationPreferences(app).timezone
#        schema = TimetableSchema(self.day_ids, timezone=tzname)
#        for day_id in self.day_ids:
#            schema[day_id] = TimetableSchemaDay(
#                [title for title, start, duration in periods])
#        schema.model = model
#        return schema
#
#    def __call__(self):
#        try:
#            data = getWidgetsData(self, self._schema)
#        except WidgetsError:
#            return self.template()
#
#        if 'CANCEL' in self.request:
#            self.request.response.redirect(
#                absoluteURL(self.context, self.request))
#        elif 'CREATE' in self.request:
#            periods = self.getPeriods()
#            if self.error:
#                return self.template()
#
#            if not periods:
#                self.error = _('You must specify at least one period.')
#                return self.template()
#
#            schema = self.createSchema(periods)
#            schema.title = data['title']
#
#            nameChooser = INameChooser(self.context)
#            name = nameChooser.chooseName('', schema)
#
#            self.context[name] = schema
#            self.request.response.redirect(
#                absoluteURL(self.context, self.request))
#
#        return self.template()



#class TimetableSchemaXMLView(BrowserView):
#    """View for ITimetableSchema"""
#
#    dows = ['Monday', 'Tuesday', 'Wednesday', 'Thursday',
#            'Friday', 'Saturday', 'Sunday']
#
#    template = ViewPageTemplateFile("templates/schema_export.pt",
#                                    content_type="text/xml; charset=UTF-8")
#
#    __call__ = template
#
#    def exceptiondayids(self):
#        result = []
#
#        for date, id in self.context.model.exceptionDayIds.items():
#            result.append({'when': str(date), 'id': id})
#
#        result.sort(lambda a, b: cmp((a['when'], a['id']),
#                                     (b['when'], b['id'])))
#        return result
#
#    def daytemplates(self):
#        items = self.context.items()
#        id = items[0][0]
#        result = []
#        for id, day in self.context.model.dayTemplates.items():
#            if id is None:
#                used = "default"
#            elif id in self.context.keys():
#                used = id
#            else:
#                used = self.dows[id]
#            periods = []
#            for period in day:
#                periods.append(
#                    {'id': None,
#                     'tstart': period.tstart.strftime("%H:%M"),
#                     'duration': period.duration.seconds / 60})
#            periods.sort()
#            for template in result:
#                if template['periods'] == periods:
#                    days = template['used'].split()
#                    days.append(used)
#                    days.sort()
#                    template['used'] = " ".join(days)
#                    break
#            else:
#                result.append({'used': used, 'periods': periods})
#
#        for date, day in self.context.model.exceptionDays.items():
#            periods = []
#            for period, slot in day:
#                periods.append(
#                    {'id': period,
#                     'tstart': slot.tstart.strftime("%H:%M"),
#                     'duration': slot.duration.seconds / 60})
#            periods.sort()
#            result.append({'used': str(date), 'periods': periods})
#
#        result.sort(lambda a, b: cmp((a['used'], a['periods']),
#                                     (b['used'], b['periods'])))
#
#        return result



class ISelectedPeriodsAddForm(Interface):
    """Form schema for ITerm add/edit views."""

    timetable = zope.schema.Choice(
        title=_("School timetable"),
        source="schooltool.timetable.browser.timetable_vocabulary",
        required=True,
    )

    first = zope.schema.Date(title=_("Apply from"))

    last = zope.schema.Date(title=_("Apply until"))


class SelectedPeriodsAddView(form.AddForm):

    template = ViewPageTemplateFile('templates/selected_periods_add.pt')
    fields = field.Fields(ISelectedPeriodsAddForm)

    _object_added = None

    buttons = button.Buttons(
        button.Button('add', title=_('Add')),
        button.Button('cancel', title=_('Cancel')))

    @property
    def owner(self):
        return IHaveSchedule(self.context)

    @property
    def term(self):
        return ITerm(self.owner, None)

    @button.handler(buttons["add"])
    def handleAdd(self, action):
        return form.AddForm.handleAdd.func(self, action)

    @button.handler(buttons["cancel"])
    def handleCancel(self, action):
        url = absoluteURL(self.context, self.request)
        self.request.response.redirect(url)

    def updateActions(self):
        super(SelectedPeriodsAddView, self).updateActions()
        self.actions['add'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')

    def create(self, data):
        timetable = data['timetable']
        schedule = SelectedPeriodsSchedule(
            timetable, data['first'], data['last'],
            title=timetable.title,
            timezone=timetable.timezone)
        return schedule

    def add(self, schedule):
        chooser = INameChooser(self.context)
        name = chooser.chooseName('', schedule)
        self.context[name] = schedule
        self._object_added = schedule

    def nextURL(self):
        if self._object_added is not None:
            return '%s/edit.html' % (
                absoluteURL(self._object_added, self.request))
        return absoluteURL(self.context, self.request)


TimetableAdd_default_first = widget.ComputedWidgetAttribute(
    lambda adapter: adapter.view.term.first,
    view=SelectedPeriodsAddView,
    field=ISelectedPeriodsAddForm['first']
    )


TimetableAdd_default_last = widget.ComputedWidgetAttribute(
    lambda adapter: adapter.view.term.last,
    view=SelectedPeriodsAddView,
    field=ISelectedPeriodsAddForm['last']
    )


class FlourishSelectedPeriodsAddView(flourish.form.AddForm,
                                     SelectedPeriodsAddView):

    legend = _("Schedule a school timetable")


SelectedPeriodsSubmitLabel = button.StaticButtonActionAttribute(
    _(u'Submit'),
    button=FlourishSelectedPeriodsAddView.buttons['add'])


class SelectedPeriodsFormValidator(validator.InvariantsValidator):

    def _formatTitle(self, object):
        if object is None:
            return None
        def dateTitle(date):
            if date is None:
                return '...'
            formatter = getMultiAdapter((date, self.request), name='mediumDate')
            return formatter()
        return u"%s (%s - %s)" % (
            object.title, dateTitle(object.first), dateTitle(object.last))

    def getOthers(self, schedule):
        container = schedule.__parent__
        others = [other for key, other in container.items()
                  if key != schedule.__name__]
        return others

#    def validateAgainstOthers(self, schedule, others):
#        pass

#    def validateAgainstTerm(self, schedule, term):
#    term_daterange = IDateRange(term)
#    if ((first is not None and first not in term_daterange) or
#        (last is not None and last not in term_daterange)):
#        raise TimetableOverflowError(
#            schema, first, last, term)

    def validateObject(self, schedule):
        #errors = super(SelectedPeriodsFormValidator, self).validateObject(schedule)
        #try:
        #    dr = DateRange(schedule.first, schedule.last)
        #
        #    others = self.getOthers()
        #
        #    term = ITerm(timetable_dict)
        #
        #    try:
        #        validateAgainstOthers(
        #            timetable.schooltt, timetable.first, timetable.last,
        #            others)
        #    except TimetableOverlapError, e:
        #        for tt in e.overlapping:
        #            errors += (Invalid(
        #                u"%s %s" % (
        #                    _("Timetable conflicts with another:"),
        #                    self._formatTitle(tt))), )
        #    try:
        #        validateAgainstTerm(
        #            timetable.schooltt, timetable.first, timetable.last,
        #            term)
        #    except TimetableOverflowError, e:
        #        errors += (Invalid(u"%s %s" % (
        #            _("Timetable does not fit in term"),
        #            self._formatTitle(term))), )
        #except ValueError, e:
        #    errors += (Invalid(_("Schedule must begin before it ends.")), )
        #except validator.NoInputData:
        #    return errors
        #return errors
        return []


class SelectedPeriodsAddFormValidator(validator.InvariantsValidator):
    def getOthers(self, schedule):
        container = self.context
        return container.values()


validator.WidgetsValidatorDiscriminators(
    SelectedPeriodsAddFormValidator,
    view=SelectedPeriodsAddView,
    schema=getSpecification(ISelectedPeriodsAddForm, force=True))


class SelectedPeriodsContent(ContentProvider):
    implements(IRenderDayTableCells)

    def __init__(self, *args, **kw):
        ContentProvider.__init__(self, *args, **kw)
        self.owner = IHaveSchedule(self.context)

    def renderCells(self, schedule, day, item):
        timetable = self.context.timetable
        if sameProxiedObjects(schedule, timetable.periods):
            if self.context.hasPeriod(item):
                return (item.title,
                        self.owner.title)
            else:
                return (item.title, '')
        return None

    def __call__(self):
        return self.template()


class ISelectedPeriodsAddTermsForm(Interface):

    timetable = zope.schema.Choice(
        title=_("School timetable"),
        source="schooltool.timetable.browser.timetable_vocabulary",
        required=True,
    )

    first = zope.schema.Choice(
        title=_("First term"),
        required=True,
        vocabulary='schooltool.course.section.linked_section_terms')

    last = zope.schema.Choice(
        title=_("Last term"),
        required=True,
        vocabulary='schooltool.course.section.linked_section_terms')


class FlourishSelectedPeriodsAddTermsView(FlourishSelectedPeriodsAddView):
    fields = field.Fields(ISelectedPeriodsAddTermsForm)

    _objects_added = None
    _objects_created = None

    buttons = button.Buttons(
        button.Button('add', title=_('Add')),
        button.Button('cancel', title=_('Cancel')))

    @property
    def owner(self):
        return IHaveSchedule(self.context)

    @button.handler(buttons["add"])
    def handleAdd(self, action):
        return form.AddForm.handleAdd.func(self, action)

    @button.handler(buttons["cancel"])
    def handleCancel(self, action):
        url = absoluteURL(self.context, self.request)
        self.request.response.redirect(url)

    def updateActions(self):
        super(SelectedPeriodsAddView, self).updateActions()
        self.actions['add'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')

    def create(self, data):
        section = self.owner
        linked = section.linked_sections
        terms = [(ITerm(s), s) for s in linked]
        terms = [(term, section) for term, section in terms
                 if (term.first >= data['first'].first and
                     term.last <= data['last'].last)]
        self._objects_created = []
        timetable = data['timetable']
        for term, section in terms:
            schedule = SelectedPeriodsSchedule(
                timetable, term.first, term.last,
                title=timetable.title,
                timezone=timetable.timezone)
            self._objects_created.append((section, schedule))
        return self._objects_created

    def add(self, schedules):
        self._objects_added = []
        for section, schedule in self._objects_created:
            schedules = IScheduleContainer(section)
            chooser = INameChooser(schedules)
            name = chooser.chooseName('', schedule)
            schedules[name] = schedule
            self._objects_added.append(schedule)

    def nextURL(self):
        if self._objects_added:
            int_ids = getUtility(IIntIds)
            also_apply = ''
            this_section = self.owner
            n = 0
            for schedule in self._objects_added:
                section = IHaveSchedule(schedule)
                if sameProxiedObjects(this_section, section):
                    continue
                also_apply = '&'.join(
                    [also_apply,
                    'section-%d=%s&schedule-%d=%s' % (
                        n, int_ids.getId(removeSecurityProxy(section)),
                        n, urllib.quote(schedule.__name__)),
                     ])
                n += 1
            if also_apply:
                also_apply = '?' + also_apply
            return '%s/edit_multi.html%s' % (
                absoluteURL(self._objects_added[0], self.request),
                also_apply)
        return absoluteURL(self.context, self.request)


SelectedPeriodsAddTermsSubmitLabel = button.StaticButtonActionAttribute(
    _(u'Submit'),
    button=FlourishSelectedPeriodsAddTermsView.buttons['add'])


TimetableAdd_default_first = widget.ComputedWidgetAttribute(
    lambda adapter: adapter.view.term.first,
    view=SelectedPeriodsAddView,
    field=ISelectedPeriodsAddForm['first']
    )


TimetableAdd_default_last = widget.ComputedWidgetAttribute(
    lambda adapter: adapter.view.term.last,
    view=SelectedPeriodsAddView,
    field=ISelectedPeriodsAddForm['last']
    )


class SelectedPeriodsAddTertiaryNav(flourish.page.TertiaryNavigationManager):
    pass


class SelectedPeriodsScheduleEditView(form.EditForm):
    implements(IRenderDayTableCells)

    template = ViewPageTemplateFile('templates/selected_periods_edit.pt')
    fields = field.Fields(ISelectedPeriodsSchedule).select(
        'first', 'last',
        'consecutive_periods_as_one')
    fields['consecutive_periods_as_one'].widgetFactory = SingleCheckBoxFieldWidget

    def getPeriodKey(self, day, period):
        return 'period.%s.%s' % (simple_form_key(day),
                                 simple_form_key(period))

    def activityTitle(self, activity_type):
        if activity_type in self.activity_vocabulary:
            term = self.activity_vocabulary.getTerm(activity_type)
            return term.title
        return None

    def renderCells(self, schedule, day, item):
        timetable = self.context.timetable
        if sameProxiedObjects(schedule, timetable.periods):
            checked = self.context.hasPeriod(item)
            key = self.getPeriodKey(day, item)
            checkbox = """
              <input class="activity" type="checkbox"
                     id="%(key)s" name="%(key)s"
                     value="%(key)s"%(checked)s></input>""" % {
                'key': key,
                'checked': checked and ' checked="checked"' or ''}
            label = """
              <label class="period" for="%(key)s">%(title)s</label>""" % {
                'key': key, 'title': item.title or ''}
            return (checkbox, label)
        if sameProxiedObjects(schedule, timetable.time_slots):
            slot = item
            return (format_time_range(slot.tstart, slot.duration),
                    self.activityTitle(slot.activity_type) or '')
        return None

    def update(self):
        self.owner = IHaveSchedule(self.context)
        self.activity_vocabulary = getActivityVocabulary(self.context)
        super(SelectedPeriodsScheduleEditView, self).update()

    def updateActions(self):
        super(SelectedPeriodsScheduleEditView, self).updateActions()
        self.actions['apply'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')

    @button.buttonAndHandler(_('Save'), name='apply')
    def handleApply(self, action):
        data, errors = self.extractData()
        if errors:
            self.status = self.formErrorsMessage
            return
        changes = self.applyChanges(data)

        schedule_changed = bool(changes)

        timetable = self.context.timetable
        for day in timetable.periods.templates.values():
            for period in day.values():
                key = self.getPeriodKey(day, period)
                selected = bool(self.request.get(key))
                scheduled = self.context.hasPeriod(period)
                if selected and not scheduled:
                    self.context.addPeriod(period)
                    schedule_changed = True
                elif not selected and scheduled:
                    self.context.removePeriod(period)
                    schedule_changed = True

        self.status = self.successMessage

        if schedule_changed:
            zope.lifecycleevent.modified(self.context)
        self.redirectToParent()

    @button.buttonAndHandler(_("Cancel"), name='cancel')
    def handle_cancel_action(self, action):
        self.redirectToParent()

    def redirectToParent(self):
        self.request.response.redirect(
            absoluteURL(self.context.__parent__,
                        self.request))

    @property
    def term(self):
        return ITerm(self.owner, None)


validator.WidgetsValidatorDiscriminators(
    SelectedPeriodsFormValidator,
    view=SelectedPeriodsScheduleEditView,
    schema=getSpecification(ITimetable, force=True))


class FlourishSelectedPeriodsScheduleEditView(flourish.page.WideContainerPage,
                                              SelectedPeriodsScheduleEditView):
    __init__ = SelectedPeriodsScheduleEditView.__init__
    update = SelectedPeriodsScheduleEditView.update


class FlourishSelectedPeriodsMultiScheduleEditView(flourish.page.WideContainerPage,
                                                   SelectedPeriodsScheduleEditView):
    __init__ = SelectedPeriodsScheduleEditView.__init__
    update = SelectedPeriodsScheduleEditView.update

    fields = field.Fields(ISelectedPeriodsSchedule).select(
        'consecutive_periods_as_one')
    fields['consecutive_periods_as_one'].widgetFactory = SingleCheckBoxFieldWidget

    def applyChangesToSchedule(self, schedule, data):
        changes = form.applyChanges(self, schedule, data)
        # ``changes`` is a dictionary; if empty, there were no changes
        if changes:
            # Construct change-descriptions for the object-modified event
            descriptions = []
            for interface, names in changes.items():
                descriptions.append(
                    zope.lifecycleevent.Attributes(interface, *names))
            # Send out a detailed object-modified event
            zope.event.notify(
                zope.lifecycleevent.ObjectModifiedEvent(schedule, *descriptions))

        schedule_changed = bool(changes)
        timetable = schedule.timetable
        for day in timetable.periods.templates.values():
            for period in day.values():
                key = self.getPeriodKey(day, period)
                selected = bool(self.request.get(key))
                scheduled = schedule.hasPeriod(period)
                if selected and not scheduled:
                    schedule.addPeriod(period)
                    schedule_changed = True
                elif not selected and scheduled:
                    schedule.removePeriod(period)
                    schedule_changed = True
        return schedule_changed

    @property
    def other_params(self):
        params = []
        n = 0
        while ('section-%d' % n in self.request and
               'schedule-%d' % n in self.request):
            params.extend([
                {'key': 'section-%d' % n, 'value': self.request.get('section-%d' % n)},
                {'key': 'schedule-%d' % n, 'value': self.request.get('schedule-%d' % n)},
                ])
            n += 1
        return params

    @property
    def other_schedules(self):
        result = []
        n = 0
        int_ids = getUtility(IIntIds)
        while ('section-%d' % n in self.request and
               'schedule-%d' % n in self.request):
            section_id = self.request.get('section-%d' % n)
            schedule_name = self.request.get('schedule-%d' % n)
            n += 1
            try:
                section_id = int(section_id)
            except ValueError:
                continue
            section = int_ids.queryObject(section_id, None)
            if section is None:
                continue
            schedules = IScheduleContainer(section)
            schedule = schedules.get(schedule_name, None)
            if schedule is None:
                continue
            result.append(schedule)
        return result

    @property
    def other_info(self):
        result = []
        for schedule in self.other_schedules:
            section = IHaveSchedule(schedule)
            term = ITerm(section)
            result.append({'owner': section, 'term': term})
        return result

    @button.buttonAndHandler(_('Save'), name='apply')
    def handleApply(self, action):
        data, errors = self.extractData()
        if errors:
            self.status = self.formErrorsMessage
            return
        changed = self.applyChangesToSchedule(self.getContent(), data)
        if changed:
            zope.lifecycleevent.modified(self.getContent())

        for schedule in self.other_schedules:
            changed = self.applyChangesToSchedule(schedule, data)
            if changed:
                zope.lifecycleevent.modified(schedule)

        self.status = self.successMessage

        self.redirectToParent()

    @button.buttonAndHandler(_("Cancel"), name='cancel')
    def handle_cancel_action(self, action):
        self.redirectToParent()


class FlourishSelectedScheduleDeleteView(FlourishConfirmDeleteView):

    def initDialog(self):
        super(FlourishSelectedScheduleDeleteView, self).initDialog()
        title = _(u'Unschedule timetable ${timetable}',
                  mapping={'timetable': self.context.timetable.title})
        self.ajax_settings['dialog']['title'] = translate(
            title, context=self.request)

    def delete(self):
        container = self.context.__parent__
        del container[self.context.__name__]

    @property
    def owner(self):
        return IHaveSchedule(self.context)


class TimetableActionsLinks(flourish.page.RefineLinksViewlet):
    """Manager for Action links in timetable views."""


class FlourishTimetableDeleteView(FlourishConfirmDeleteView):

    def initDialog(self):
        super(FlourishTimetableDeleteView, self).initDialog()
        title = _(u'Delete ${timetable}',
                  mapping={'timetable': self.context.title})
        self.ajax_settings['dialog']['title'] = translate(
            title, context=self.request)

    def delete(self):
        container = ITimetableContainer(self.context)
        del container[self.context.__name__]


class FlourishTimetableMakeDefaultView(flourish.form.DialogForm, form.EditForm):
    dialog_submit_actions = ('apply',)
    dialog_close_actions = ('cancel',)
    label = None

    def initDialog(self):
        super(FlourishTimetableMakeDefaultView, self).initDialog()
        title = _(u'Make ${timetable} default',
                  mapping={'timetable': self.context.title})
        self.ajax_settings['dialog']['title'] = translate(
            title, context=self.request)

    @button.buttonAndHandler(_("Apply"), name='apply')
    def handleApply(self, action):
        container = ITimetableContainer(self.context)
        container.default = removeSecurityProxy(self.context)
        self.ajax_settings['dialog'] = 'close'

    @button.buttonAndHandler(_("Cancel"))
    def handleCancel(self, action):
        pass

    def updateActions(self):
        super(FlourishTimetableMakeDefaultView, self).updateActions()
        self.actions['apply'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')



#
#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2009 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Contact browser views.
"""
import urllib

from zope.security.checker import canAccess
from zope.schema import getFieldsInOrder
from zope.interface import implements, Interface
from zope.catalog.interfaces import ICatalog
from zope.component import getMultiAdapter
from zope.component import adapts
from zope.component import getUtility
from zope.schema import getFieldNamesInOrder, TextLine, Text
from zope.security.proxy import removeSecurityProxy
from zope.traversing.browser.interfaces import IAbsoluteURL
from zope.traversing.browser import absoluteURL
from zope.publisher.interfaces.browser import IBrowserRequest
from zope.publisher.browser import BrowserView
from zope.container.interfaces import INameChooser
from zope.browserpage.viewpagetemplatefile import ViewPageTemplateFile
from zope.intid.interfaces import IIntIds
from zope.i18n import translate

from z3c.form import form, subform, field, button
from z3c.form.interfaces import DISPLAY_MODE

from schooltool.app.browser.app import ActiveSchoolYearContentMixin
from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.app.catalog import buildQueryString
from schooltool.contact.interfaces import IContactable
from schooltool.contact.interfaces import IContactContainer
from schooltool.contact.interfaces import IContactPersonInfo
from schooltool.contact.interfaces import IAddress, IPhoto
from schooltool.contact.interfaces import IUniqueFormKey
from schooltool.contact.contact import ContactPersonInfo
from schooltool.contact.interfaces import IContact
from schooltool.contact.contact import Contact
from schooltool.person.interfaces import IPerson
from schooltool.person.interfaces import IPersonFactory
from schooltool.email.interfaces import IEmailUtility
from schooltool.email.mail import Email
from schooltool.skin import flourish
from schooltool.skin.containers import TableContainerView
from schooltool.relationship.relationship import IRelationshipLinks
from schooltool.common.inlinept import InlineViewPageTemplate
from schooltool.contact.contact import URIPerson, URIContact
from schooltool.contact.contact import URIContactRelationship
from schooltool.contact.interfaces import IContactPerson
from schooltool.contact.interfaces import IEmails, IPhones, ILanguages
from schooltool import table

from schooltool.common import SchoolToolMessage as _


class ContactContainerAbsoluteURLAdapter(BrowserView):

    adapts(IContactContainer, IBrowserRequest)
    implements(IAbsoluteURL)

    def __str__(self):
        app = ISchoolToolApplication(None)
        url = str(getMultiAdapter((app, self.request), name='absolute_url'))
        return url + '/contacts'

    __call__ = __str__


class FlourishContactContainerView(flourish.page.Page):
    """A flourish Contact Container view."""

    content_template = InlineViewPageTemplate('''
      <div tal:content="structure context/schooltool:content/ajax/table" />
    ''')

    @property
    def done_link(self):
        app = ISchoolToolApplication(None)
        return absoluteURL(app, self.request) + '/manage'


class ContactAddView(form.AddForm):
    """Contact add form for basic contact."""

    label = _("Add new contact")
    template = ViewPageTemplateFile('templates/contact_add.pt')
    fields = field.Fields(IContact)
    formErrorsMessage = _('Please correct the marked fields below.')

    def __init__(self, *args, **kw):
        super(ContactAddView, self).__init__(*args, **kw)
        self.subforms = []

    def updateActions(self):
        super(ContactAddView, self).updateActions()
        self.actions['add'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')

    @button.buttonAndHandler(_('Add'), name='add')
    def handleAdd(self, action):
        data, errors = self.extractData()
        if errors:
            self.status = self.formErrorsMessage
            return
        obj = self.createAndAdd(data)
        if obj is not None:
            # mark only as finished if we get the new object
            self._finishedAdd = True

    def create(self, data):
        contact = Contact()
        form.applyChanges(self, contact, data)
        return contact

    def nextURL(self):
        return absoluteURL(self.context, self.request)

    def add(self, contact):
        """Add `contact` to the container.

        Uses the username of `contact` as the object ID (__name__).
        """
        name = INameChooser(self.context).chooseName('', contact)
        self.context[name] = contact
        return contact

    @button.buttonAndHandler(_("Cancel"))
    def handle_cancel_action(self, action):
        self.request.response.redirect(self.nextURL())


class FlourishContactAddView(flourish.page.Page, ContactAddView):
    label = None

    def update(self):
        self.buildFieldsetGroups()
        ContactAddView.update(self)

    def render(self):
        if self._finishedAdd:
            self.request.response.redirect(self.nextURL())
            return ""
        return super(FlourishContactAddView, self).render()

    def makeRows(self, fields, cols=1):
        rows = []
        while fields:
            rows.append(fields[:cols])
            fields = fields[cols:]
        return rows

    def makeFieldSet(self, fieldset_id, legend, fields, cols=1):
        result = {
            'id': fieldset_id,
            'legend': legend,
            }
        result['rows'] = self.makeRows(fields, cols)
        return result

    def fieldsets(self):
        result = []
        for fieldset_id in self.fieldset_order:
            legend, fields = self.fieldset_groups[fieldset_id]
            result.append(self.makeFieldSet(
                    fieldset_id, legend, list(fields)))
        return result

    def buildFieldsetGroups(self):
        self.fieldset_groups = {
            'full_name': (
                _('Full Name'),
                ['prefix', 'first_name', 'middle_name', 'last_name',
                 'suffix']),
            'details': (
                _('Details'), ['photo']),
            'address': (
                _('Address'),
                ['address_line_1', 'address_line_2', 'city', 'state',
                 'country', 'postal_code']),
            'contact_information': (
                _('Contact Information'),
                ['home_phone', 'work_phone', 'mobile_phone', 'email',
                 'other_1', 'other_2',
                 'language']),
            }
        self.fieldset_order = (
            'full_name', 'details', 'address', 'contact_information')


SubmitLabel = button.StaticButtonActionAttribute(
    _('Submit'),
    button=FlourishContactAddView.buttons['add'])


class ContactPersonInfoSubForm(subform.EditSubForm):
    """Form for editing additional person's contact information."""

    template = ViewPageTemplateFile('templates/contactperson_subform.pt')
    fields = field.Fields(IContactPersonInfo)
    prefix = 'person_contact'

    changed = False

    @property
    def person(self):
        return self.context.__parent__

    @button.handler(form.EditForm.buttons['apply'])
    def handleApply(self, action):
        data, errors = self.widgets.extract()
        if errors:
            # XXX: we don't handle errors for now
            pass
        content = self.getContent()
        changed = form.applyChanges(self, content, data)
        self.changed = bool(changed)

    @button.handler(ContactAddView.buttons['add'])
    def handleAdd(self, action):
        # Pretty, isn't it?
        self.handleApply.func(self, action)


class PersonContactAddView(ContactAddView):
    """Contact add form that assigns the contact to a person."""

    form.extends(ContactAddView)

    @property
    def label(self):
        return _("Add new contact for ${person}",
                 mapping={'person': self.context.title})

    def update(self):
        # Create contact_info object for adding
        self.contact_info = ContactPersonInfo()
        # Add person contact.
        super(PersonContactAddView, self).update()
        # Update contact info.
        self.subforms = [
            ContactPersonInfoSubForm(self.contact_info, self.request, self),
            ]
        for subform in self.subforms:
            subform.update()

    def add(self, contact):
        """Add `contact` to the container and assign it to the person."""
        contact_container = IContactContainer(ISchoolToolApplication(None))
        name = INameChooser(contact_container).chooseName('', contact)
        contact_container[name] = contact

        context = removeSecurityProxy(self.context)
        self.contact_info.__parent__ = context
        IContactable(context).contacts.add(contact, self.contact_info)
        return contact


class FlourishPersonContactAddView(FlourishContactAddView,
                                   PersonContactAddView):

    label = None

    add = PersonContactAddView.add

    def update(self):
        self.buildFieldsetGroups()
        PersonContactAddView.update(self)

    def nextURL(self):
        base_url = absoluteURL(self.context, self.request)
        return "%s/@@manage_contacts.html?%s" % (
            base_url,
            urllib.urlencode([('SEARCH_TITLE',
                               self.context.last_name.encode("utf-8"))]))


class ContactEditView(form.EditForm):
    """Edit form for basic contact."""
    form.extends(form.EditForm)
    template = ViewPageTemplateFile('templates/contact_add.pt')
    fields = field.Fields(IContact)

    formErrorsMessage = _('Please correct the marked fields below.')

    @button.buttonAndHandler(_("Cancel"))
    def handle_cancel_action(self, action):
        url = absoluteURL(self.context, self.request)
        self.request.response.redirect(url)

    def update(self):
        super(ContactEditView, self).update()
        self.subforms = []
        for relationship_info in self.context.persons.relationships:
            subform = ContactPersonInfoSubForm(
                relationship_info.extra_info, self.request, self)
            # XXX: should also apply at least urllib.quote here:
            prefix = unicode(relationship_info.target.__name__).encode('punycode')
            subform.prefix += '.%s' % prefix
            subform.update()
            # One more hack.
            if subform.changed and self.status == self.noChangesMessage:
                self.status = self.successMessage
            self.subforms.append(subform)

    def updateActions(self):
        super(ContactEditView, self).updateActions()
        self.actions['apply'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')

    @property
    def label(self):
        return _(u'Change contact information for ${person}',
                 mapping={'person': self.context.title})


class FlourishContactEditView(flourish.page.Page,
                              ContactEditView):

    form.extends(ContactEditView)

    label = None

    def update(self):
        self.buildFieldsetGroups()
        ContactEditView.update(self)

    @button.handler(ContactEditView.buttons['apply'])
    def handle_apply(self, action):
        # Pretty, isn't it?
        self.handleApply.func(self, action)
        # XXX: hacky sucessful submit check
        if (self.status == self.successMessage or
            self.status == self.noChangesMessage):
            self.request.response.redirect(self.nextURL())

    def nextURL(self):
        if 'person_id' in self.request:
            person_id = self.request['person_id']
            app = ISchoolToolApplication(None)
            persons = app['persons']
            if person_id in persons:
                return absoluteURL(persons[person_id], self.request)
        return absoluteURL(self.context, self.request)

    @button.handler(ContactEditView.buttons['cancel'])
    def handle_cancel_action(self, action):
        self.request.response.redirect(self.nextURL())

    def makeRows(self, fields, cols=1):
        rows = []
        while fields:
            rows.append(fields[:cols])
            fields = fields[cols:]
        return rows

    def makeFieldSet(self, fieldset_id, legend, fields, cols=1):
        result = {
            'id': fieldset_id,
            'legend': legend,
            }
        result['rows'] = self.makeRows(fields, cols)
        return result

    def buildFieldsetGroups(self):
        self.fieldset_groups = {
            'full_name': (
                _('Full Name'),
                ['prefix', 'first_name', 'middle_name', 'last_name',
                 'suffix']),
            'details': (
                _('Details'), ['photo']),
            'address': (
                _('Address'),
                ['address_line_1', 'address_line_2', 'city', 'state',
                 'country', 'postal_code']),
            'contact_information': (
                _('Contact Information'),
                ['home_phone', 'work_phone', 'mobile_phone', 'email',
                 'other_1', 'other_2',
                 'language']),
            }
        self.fieldset_order = (
            'full_name', 'details', 'address', 'contact_information')

    def fieldsets(self):
        result = []
        for fieldset_id in self.fieldset_order:
            legend, fields = self.fieldset_groups[fieldset_id]
            result.append(self.makeFieldSet(
                    fieldset_id, legend, list(fields)))
        return result


class ContactView(form.DisplayForm):
    """Display form for basic contact."""
    template = ViewPageTemplateFile('templates/contact_view.pt')
    fields = field.Fields(IContact)

    def relationships(self):
        return [relationship_info.extra_info
                for relationship_info in self.context.persons.relationships]

    def __call__(self):
        self.update()
        return self.render()


class FlourishContactView(flourish.page.Page):

    pass


class FlourishBoundContactDetails(flourish.form.FormContent):

    fields = field.Fields(IContact).omit(*(
        list(IPhoto) +
        list(IContactPerson)
        ))
    mode = DISPLAY_MODE

    @property
    def has_data(self):
        return any([widget.value for widget in self.widgets.values()])

    def relationships(self):
        return [relationship_info.extra_info
                for relationship_info in self.context.persons.relationships]

    @property
    def canModify(self):
        return canAccess(self.context.__parent__, '__delitem__')


class FlourishContactDetails(flourish.form.FormViewlet):

    template = ViewPageTemplateFile('templates/f_contact_view.pt')
    fields = field.Fields(IContact).omit('photo')
    mode = DISPLAY_MODE

    def relationships(self):
        return [relationship_info.extra_info
                for relationship_info in self.context.persons.relationships]

    @property
    def canModify(self):
        return canAccess(self.context.__parent__, '__delitem__')


class ContactContainerView(TableContainerView):
    """A Contact Container view."""

    __used_for__ = IContactContainer

    delete_template = ViewPageTemplateFile('templates/contacts_delete.pt')
    index_title = _("Contact index")

    @property
    def itemsToDelete(self):
        return sorted(
            TableContainerView._listItemsForDeletion(self),
            key=lambda obj: '%s %s' % (obj.last_name, obj.first_name))

    def columnsBefore(self):
        if self.canModify():
            return [
                table.column.DependableCheckboxColumn(
                    prefix="delete",
                    name='delete_checkbox',
                    title=u'',
                    id_getter=IUniqueFormKey,
                    show_disabled=False)]
        return []

    def setUpTableFormatter(self, formatter):
        columns_before = self.columnsBefore()
        formatter.setUp(columns_before=columns_before)

    def listIdsForDeletion(self):
        int_ids = getUtility(IIntIds)
        return [int_ids.getObject(intid).__name__
                for intid in self.listIntIdsForDeletion()]

    def listFormKeysForDeletion(self):
        catalog = ICatalog(self.context)
        index = catalog['form_keys']
        return ['delete.%s' % index.documents_to_values[intid]
                for intid in self.listIntIdsForDeletion()]

    def listIntIdsForDeletion(self):
        catalog = ICatalog(self.context)
        index = catalog['form_keys']
        return [intid for intid in catalog.extent
                if 'delete.%s' % index.documents_to_values[intid] in self.request]


def format_street_address(item, formatter):
    address_parts = []
    for attribute in getFieldNamesInOrder(IAddress):
        address_part = getattr(item, attribute, None)
        if address_part is not None:
            address_parts.append(address_part)
    return ", ".join(address_parts)


def contact_table_columns():
    return getUtility(IPersonFactory).columns()


class ContactTableFormatter(table.catalog.IndexedTableFormatter):

    columns = lambda self: contact_table_columns()

    def sortOn(self):
        return getUtility(IPersonFactory).sortOn()


class FlourishContactTableFormatter(ContactTableFormatter):

    def makeFormatter(self):
        formatter = ContactTableFormatter.makeFormatter(self)
        formatter.cssClasses['table'] = 'contacts-table relationships-table'
        return formatter

    def sortOn(self):
        return getUtility(IPersonFactory).sortOn()


class ContactTable(table.ajax.IndexedTable, FlourishContactTableFormatter):
    columns = FlourishContactTableFormatter.columns
    sortOn = FlourishContactTableFormatter.sortOn


class ContactFilterWidget(table.catalog.IndexedFilterWidget):

    template = ViewPageTemplateFile('templates/filter.pt')
    parameters = ['SEARCH_FIRST_NAME', 'SEARCH_LAST_NAME']

    def filter(self, items):
        if 'CLEAR_SEARCH' in self.request:
            for parameter in self.parameters:
                self.request.form[parameter] = ''
            return items

        catalog = self.catalog

        if 'SEARCH_FIRST_NAME' in self.request:
            # XXX: applying normalized catalog queries would be nicer
            fn_idx = catalog['first_name']
            searchstr = self.request['SEARCH_FIRST_NAME'].lower()
            items = [item for item in items
                     if searchstr in fn_idx.documents_to_values[item['id']].lower()]

        if 'SEARCH_LAST_NAME' in self.request:
            # XXX: applying normalized catalog queries would be nicer
            ln_idx = catalog['last_name']
            searchstr = self.request['SEARCH_LAST_NAME'].lower()
            items = [item for item in items
                     if searchstr in ln_idx.documents_to_values[item['id']].lower()]

        return items

    def active(self):
        for parameter in self.parameters:
            if parameter in self.request:
                return True
        return False

    def quote(self, param):
        return urllib.quote(unicode(param).encode('UTF-8'))

    def extra_url(self):
        url = ""
        for parameter in self.parameters:
            if parameter in self.request:
                url += '&%s=%s' % (parameter, self.quote(self.request.get(parameter)))
        return url


class FlourishContactFilterWidget(ContactFilterWidget):

    template = ViewPageTemplateFile('templates/f_contact_filter.pt')

    parameters = ['SEARCH_TITLE']

    search_title_id = 'SEARCH_TITLE'

    def filter(self, items):
        if self.search_title_id in self.request:
            search_title = self.request[self.search_title_id]
            query = buildQueryString(search_title)
            if query:
                catalog = self.catalog
                result = catalog['text'].apply(query)
                items = [item for item in items
                         if item['id'] in result]
        return items


class ContactTableFilter(table.ajax.IndexedTableFilter,
                         FlourishContactFilterWidget):

    template = ViewPageTemplateFile('templates/f_contact_table_filter.pt')
    title = _('First name and/or last name')

    @property
    def search_title_id(self):
        return self.manager.html_id+"-title"

    @property
    def parameters(self):
        return (self.search_title_id, )

    active = FlourishContactFilterWidget.active
    extra_url = FlourishContactFilterWidget.extra_url

    def filter(self, results):
        if self.ignoreRequest:
            return results
        return FlourishContactFilterWidget.filter(self, results)

class ContactBackToContainerViewlet(object):
    @property
    def link(self):
        container = IContactContainer(ISchoolToolApplication(None))
        return absoluteURL(container, self.request)


class SendEmailActionViewlet(object):
    @property
    def link(self):
        utility = getUtility(IEmailUtility)
        if utility.enabled() and self.context.email:
            return absoluteURL(self.context, self.request) + '/sendEmail.html'
        return ''


class ISendEmailForm(Interface):

    from_address = TextLine(title=_(u'From'), required=False)
    to_addresses = TextLine(title=_(u'To'), required=False)
    subject = TextLine(title=_(u'Subject'))
    body = Text(title=_(u'Body'))


class SendEmailFormAdapter(object):

    implements(ISendEmailForm)
    adapts(IContact)

    def __init__(self, context):
        self.context = context


class SendEmailView(form.Form):

    template = ViewPageTemplateFile('templates/email_form.pt')
    label = _('Send Email')
    fields = field.Fields(ISendEmailForm)

    @button.buttonAndHandler(_('Send'))
    def handle_send_action(self, action):
        data, errors = self.extractData()
        if errors:
            self.status = _('There were some errors.')
            return
        body = data['body']
        subject = data['subject']
        email = Email(self.from_address, self.to_addresses, body, subject)
        utility = getUtility(IEmailUtility)
        success = utility.send(email)
        url = absoluteURL(self.context, self.request)
        self.request.response.redirect(url)

    @button.buttonAndHandler(_('Cancel'))
    def handle_cancel_action(self, action):
        url = absoluteURL(self.context, self.request)
        self.request.response.redirect(url)

    def updateActions(self):
        super(SendEmailView, self).updateActions()
        self.actions['cancel'].addClass('button-cancel')
        utility = getUtility(IEmailUtility)
        if utility.enabled() and self.context.email:
            self.actions['send'].addClass('button-ok')
        else:
            self.actions['send'].mode = 'display'

    def update(self):
        user = IPerson(self.request.principal)
        super(SendEmailView, self).update()
        if not IContact(user).email:
            url = absoluteURL(self.context, self.request) + \
                  '/noTeacherEmail.html'
            self.request.response.redirect(url)
            return
        self.updateDisplayWidgets()

    def updateDisplayWidgets(self):
        self.widgets['from_address'].mode = 'display'
        self.widgets['to_addresses'].mode = 'display'
        self.widgets['from_address'].value = "%s <%s>" % (self.sender,
                                                          self.from_address)
        self.widgets['to_addresses'].value = "%s <%s>" % (self.recipient,
                                                          ''.join(self.to_addresses))

    @property
    def from_address(self):
        user = IPerson(self.request.principal, None)
        if user is None:
            return u''
        return IContact(user).email

    @property
    def to_addresses(self):
        return [self.context.email]

    @property
    def sender(self):
        user = IPerson(self.request.principal, None)
        if user is None:
            return u''
        return user.title

    @property
    def recipient(self):
        return self.context.title


class NoTeacherEmailView(BrowserView):

    __call__ = ViewPageTemplateFile('templates/noteacheremail.pt')


class BoundContactPersonActionViewlet(object):

    @property
    def person(self):
        return IPerson(self.context)


class FlourishContactsViewlet(flourish.viewlet.Viewlet):
    """A viewlet showing contacts of a person"""

    template = ViewPageTemplateFile('templates/f_contactsViewlet.pt')

    @property
    def bound_contact(self):
        return IContact(self.context)

    @property
    def getContacts(self):
        contacts = IContactable(removeSecurityProxy(self.context)).contacts
        return [self.buildInfo(contact) for contact in contacts]

    # XXX: copied from relationship.py to avoid circular dependency
    def get_relationship_title(self, contact):
        try:
            links = IRelationshipLinks(self.context)
            link = links.find(
                URIPerson, contact, URIContact, URIContactRelationship)
        except ValueError:
            return u''
        return link.extra_info.getRelationshipTitle()

    def buildInfo(self, contact):
        return {
            'link': absoluteURL(contact, self.request),
            'relationship': self.get_relationship_title(contact),
            'name': " ".join(self._extract_attrs(
                contact, IContactPerson)),
            'address': ", ".join(self._extract_attrs(contact, IAddress)),
            'emails': ", ".join(self._extract_attrs(contact, IEmails)),
            'phones': list(self._extract_attrs(contact, IPhones, add_title=True)),
            'languages': ", ".join(self._extract_attrs(contact, ILanguages)),
            'obj': contact,
            }

    def _extract_attrs(self, contact, interface,
                       conjunctive=", ", add_title=False):
        parts = []
        for name, field in getFieldsInOrder(interface):
            part = getattr(contact, name, None)
            part = part and part.strip() or part
            if not part:
                continue
            if add_title:
                parts.append("%s (%s)" % (part, field.title))
            else:
                parts.append(part)
        return parts

    @property
    def canModify(self):
        return canAccess(self.context.__parent__, '__delitem__')

    def makeRows(self, contact):
        rows = []
        fields = field.Fields(IAddress, IEmails, IPhones, ILanguages)
        for attr in fields:
            value = getattr(contact, attr)
            if value:
                label = fields[attr].field.title
                rows.append(self.makeRow(label, value))
        return rows

    def makeRow(self, attr, value):
        if value is None:
            value = u''
        return {
            'label': attr,
            'value': unicode(value),
            }

    @property
    def manage_link(self):
        base_url = absoluteURL(self.context, self.request)
        return "%s/@@manage_contacts.html?%s" % (
            base_url,
            urllib.urlencode([('SEARCH_TITLE',
                               self.context.last_name.encode("utf-8"))]))


class ContactsLinks(flourish.page.RefineLinksViewlet):
    """Links for manage contact page"""


class PersonAsContactLinkViewlet(flourish.page.LinkIdViewlet):

    @property
    def title(self):
        person = self.context
        return _('${person_full_name} as Contact',
                 mapping={'person_full_name': person.title})


class FlourishManageContactsOverview(flourish.page.Content,
                                     ActiveSchoolYearContentMixin):

    body_template = ViewPageTemplateFile(
        'templates/f_manage_contacts_overview.pt')

    @property
    def contacts(self):
        app = ISchoolToolApplication(None)
        contacts = IContactContainer(app)
        return contacts

    @property
    def total(self):
        catalog = ICatalog(self.contacts)
        return len(catalog.extent)

    def contacts_url(self):
        return self.url_with_schoolyear_id(self.context, view_name='contacts')


class ContactActionsLinks(flourish.page.RefineLinksViewlet):
    """Contact actions links viewlet."""


class FlourishContactDeleteView(flourish.form.DialogForm, form.EditForm):
    """View used for deleting a contact."""

    dialog_submit_actions = ('apply',)
    dialog_close_actions = ('cancel',)
    label = None

    def initDialog(self):
        super(FlourishContactDeleteView, self).initDialog()
        contact = self.context
        self.ajax_settings['dialog']['title'] = translate(
            _(u'Delete ${contact_full_name}',
              mapping={'contact_full_name': contact.title}),
              context=self.request)

    @button.buttonAndHandler(_("Delete"), name='apply')
    def handleDelete(self, action):
        url = '%s/delete.html?delete.%s&CONFIRM' % (
            absoluteURL(self.context.__parent__, self.request),
            self.context.__name__)
        self.request.response.redirect(url)
        # We never have errors, so just close the dialog.
        self.ajax_settings['dialog'] = 'close'

    @button.buttonAndHandler(_("Cancel"))
    def handleCancel(self, action):
        pass

    def updateActions(self):
        super(FlourishContactDeleteView, self).updateActions()
        self.actions['apply'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')


class PhotoView(flourish.widgets.ImageView):
    attribute = "photo"



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2011 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from zope.app.generations.utility import findObjectsProviding
from zope.annotation.interfaces import IAnnotatable, IAnnotations
from zope.component import getUtility
from zope.intid.interfaces import IIntIds

from schooltool.course.interfaces import ISection
from schooltool.generations.evolve36.helper import assert_not_broken
from schooltool.generations.evolve36.helper import BuildContext
from schooltool.generations.evolve36.timetable_builders import (
    APP_TIMETABLES_KEY)
from schooltool.timetable.schedule import ScheduleContainer
from schooltool.timetable.timetable import SelectedPeriodsSchedule
from schooltool.timetable.app import SchoolToolSchedules


TIMETABLE_DICT_KEY = 'schooltool.timetable.timetables'
APP_SCHEDULES_KEY = 'schooltool.timetable.schedules'


class ScheduleBuilder(object):

    data = None
    store_data = ('__name__', 'first', 'last',
                   'timezone', 'term',
                   'consecutive_periods_as_one')
    schema_id = None
    year_int_id = None

    def read(self, timetable, context):
        self.data = {}
        for name in self.store_data:
            self.data[name] = getattr(timetable, name, None)

        for value in self.data.values():
            assert_not_broken(value)

        schema = timetable.schooltt
        assert_not_broken(schema)
        self.schema_id = schema.__name__
        self.year_int_id = int(schema.__parent__.__name__)

        self.selected_period_keys = [
            (self.year_int_id, self.schema_id, day_id, period_id)
            for day_id, period_id, activity in timetable.activities()]

    def build(self, container, context):
        timetables = context.shared.timetable_root[unicode(self.year_int_id)]
        timetable = timetables[self.schema_id]

        schedule = SelectedPeriodsSchedule(
            timetable, self.data['first'], self.data['last'],
            timezone=self.data['timezone'] or 'UTC')

        result = BuildContext(
            schedule=schedule,
            unique_key=(self.data['term'],
                        context.owner,
                        self.data['__name__'])
            )

        schedule.consecutive_periods_as_one = \
            self.data['consecutive_periods_as_one']

        for key in self.selected_period_keys:
            period = context.shared.period_map[key]
            schedule.addPeriod(period)

        container[self.data['__name__']] = schedule
        return result(term=self.data['term'])


class SchedulesBuilder(object):
    builders = None

    owner = None

    def read(self, timetable_dict, context):
        self.builders = []
        self.owner = timetable_dict.__parent__
        for key, timetable in sorted(timetable_dict.items()):
            assert_not_broken(timetable)
            builder = ScheduleBuilder()
            builder.read(timetable, context(owner=self.owner, key=key))
            self.builders.append(builder)

    def clean(self, context):
        annotations = IAnnotations(self.owner, None)
        if TIMETABLE_DICT_KEY in annotations:
            del annotations[TIMETABLE_DICT_KEY]

    def build(self, schedule_root, context):
        result = BuildContext(schedule_map={})
        if not ISection.providedBy(self.owner):
            return result(schedules=None)

        owner_int_id = getUtility(IIntIds).getId(self.owner)
        key = unicode(owner_int_id)
        container = schedule_root[key] = ScheduleContainer()

        for builder in self.builders:
            built = builder.build(
                container, context(schedule_root=schedule_root,
                                   owner=self.owner))
            result.schedule_map[built.unique_key] = built.schedule

        return result(schedules=container)


class AppSchedulesBuilder(object):
    builders = None

    def read(self, app, context):
        self.builders = []
        # Find objects that may have provided IOwnTimetables
        candidates = findObjectsProviding(app, IAnnotatable)
        for candidate in candidates:
            assert_not_broken(candidate)
            annotations = IAnnotations(candidate, None)
            if annotations is None:
                continue
            timetable_dict = annotations.get(TIMETABLE_DICT_KEY)
            if timetable_dict is None:
                continue
            assert_not_broken(timetable_dict)
            builder = SchedulesBuilder()
            builder.read(timetable_dict, context(app=app))
            self.builders.append(builder)

    def clean(self, app, context):
        for builder in self.builders:
            builder.clean(context)

    def build(self, app, context):

        if APP_SCHEDULES_KEY not in app:
            app[APP_SCHEDULES_KEY] = SchoolToolSchedules()
        schedule_root = app[APP_SCHEDULES_KEY]

        result = BuildContext(schedule_map={})

        for builder in self.builders:
            built = builder.build(schedule_root, context(app=app))
            result.schedule_map.update(built.schedule_map)

        return result(schedule_root=schedule_root)



#
#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2009 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Contact interfaces
"""
import zope.schema
from zope.annotation.interfaces import IAttributeAnnotatable
from zope.container.constraints import contains
from zope.container.constraints import containers
from zope.location.interfaces import IContained
from zope.interface import Attribute
from zope.interface import Interface
from zope.schema import TextLine
from zope.schema.interfaces import IContainer

from schooltool.skin import flourish
from schooltool.common import SchoolToolMessage as _
from schooltool.app.utils import vocabulary


class IContactPerson(Interface):
    """Name of a person."""

    prefix = TextLine(title=_(u"Prefix"), required=False)

    first_name = TextLine(title=_(u"First name"))

    middle_name = TextLine(title=_(u"Middle name"), required=False)

    last_name = TextLine(title=_(u"Last name"))

    suffix = TextLine(title=_(u"Suffix"), required=False)

    title = Attribute('Full name')


class IPhoto(Interface):

    photo = flourish.fields.Image(
        title=_('Photo'),
        size=(99,132),
        format='JPEG',
        required=False)


class IAddress(Interface):

    address_line_1 = TextLine(title=_(u"Address line 1"), required=False)

    address_line_2 = TextLine(title=_(u"Address line 2"), required=False)

    city = TextLine(title=_(u"City"), required=False)

    state = TextLine(title=_(u"State"), required=False)

    country = TextLine(title=_(u"Country"), required=False)

    postal_code = TextLine(title=_(u"Postal code"), required=False)


class IEmails(Interface):

    email = TextLine(title=_(u"Email"), required=False)


class IPhones(Interface):

    home_phone = TextLine(title=_(u"Home phone"), required=False)

    work_phone = TextLine(title=_(u"Work phone"), required=False)

    mobile_phone = TextLine(title=_(u"Mobile phone"), required=False)


class IOther(Interface):

    other_1 = TextLine(title=_(u"Other 1"), required=False)

    other_2 = TextLine(title=_(u"Other 2"), required=False)


class ILanguages(Interface):

    language = TextLine(title=_(u"Language"), required=False)


class IContactInformation(IPhoto, IAddress, IEmails, IPhones, IOther, ILanguages):
    """Collection of parts that define the contact information."""


class IContact(IContactPerson, IContactInformation, IAttributeAnnotatable):

    persons = Attribute("Persons attached to this contact (see IRelationshipProperty)")


# XXX: naming is becoming confusing.  Rename to IContactRelationshipInfo.
class IContactPersonInfo(Interface):
    """Information about a contact - person relationship."""

    __parent__ = Attribute("""The person.""")

    relationship = zope.schema.Choice(
        title=_(u"Relationship"),
        description=_("Contact's relationship with the person"),
        vocabulary=vocabulary([
            ('parent', _("Parent")),
            ('step_parent', _("Step-parent")),
            ('foster_parent', _("Foster parent")),
            ('guardian', _("Guardian")),
            ('sibling', _("Sibling")),
            ]),
        required=False)

    def getRelationshipTitle():
        """Return a title from the vocabulary for the selected relationship."""


class IContactContainer(IContainer):
    """Container of contacts."""

    contains(IContact)


class IContactContained(IContact, IContained):
    """Contact contained in an IContactContainer."""

    containers(IContactContainer)


class IContactable(Interface):
    """Object that can have contacts."""

    contacts = Attribute("Contacts (see IRelationshipProperty)")


class IUniqueFormKey(Interface):
    """Adapt a contact to this interface to obtain a unique form key."""




#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2012 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
SchoolTool flourish zcml directives.
"""
import zope.browserpage.metadirectives
import zope.component.zcml
import zope.configuration.fields
import zope.schema
import zope.security.checker
import zope.viewlet.metadirectives
from zope.interface import Interface
from zope.publisher.interfaces.browser import IBrowserRequest
from zope.publisher.interfaces.browser import IBrowserView

from schooltool.skin.flourish.zcml import IRenderOverrides
from schooltool.skin.flourish.zcml_content import subclass_content, template_specs
from schooltool.skin.flourish.zcml_content import handle_interfaces
from schooltool.skin.flourish.zcml_content import handle_security
from schooltool.skin.flourish.zcml_content import TemplatePath
from schooltool.skin.flourish import interfaces
from schooltool.skin.flourish.report import PDFPage


class IPDFPageDirective(zope.browserpage.metadirectives.IPagesDirective,
                    IRenderOverrides):

    name = zope.schema.TextLine(
        title=u"The name of the PDF (view)",
        required=True,
        )

    title = zope.configuration.fields.MessageID(
        title=u"Title of this PDF",
        required=False,
        )

    author = zope.configuration.fields.MessageID(
        title=u"Author of this PDF",
        description=u"""
        A very short description of this PDF.
        """,
        required=False,
        )

    template = TemplatePath(
        title=u"Main template",
        description=u"""
        Change the main template that renders everything.
        """,
        required=False,
        )

    content_template = TemplatePath(
        title=u"Template for the main PDF content.",
        description=u"""
        Set template that renders main content for this PDF.
        """,
        required=False,
        )


# Arbitrary keys and values are allowed to be passed to the manager.
IPDFPageDirective.setTaggedValue('keyword_arguments', True)


def pdf(_context, name, permission,
        for_=Interface, layer=interfaces.IFlourishLayer,
        title=None, author=None,
        template=None, content_template=None,
        class_=PDFPage,
        update='update', render='render',
        allowed_interface=(), allowed_attributes=(),
        **kwargs
        ):

    forward_methods = {
        'update': update,
        'render': render,
        }

    # BBB: add index to ease porting from old style views
    if (IBrowserView.implementedBy(class_) and
        getattr(class_, 'index', None) is None):
        forward_methods['index'] = render

    if not interfaces.IPDFPage.implementedBy(class_):
        class_ = type(class_.__name__, (class_, PDFPage), {})

    allowed_interface = (tuple(allowed_interface) +
                         (interfaces.IPDFPage, ))

    class_dict = dict(kwargs)
    class_dict['__name__'] = name

    if title is not None:
        class_dict['title'] = title
    if author is not None:
        class_dict['author'] = author

    # XXX: raise ConfigurationError if class_ is PDFPage and
    #      no templates specified

    templates = template_specs({
        'template': template,
        'content_template': content_template,
        }, content_type='xml')

    class_ = subclass_content(
        class_, name,
        forward_methods,
        templates,
        class_dict,
        )

    handle_interfaces(_context, (for_,))
    handle_interfaces(_context, allowed_interface)

    handle_security(class_, permission,
                    allowed_interface, allowed_attributes)

    _context.action(
        discriminator=('view', for_, name, IBrowserRequest, layer),
        callable=zope.component.zcml.handler,
        args=('registerAdapter',
              class_, (for_, layer), Interface, name, _context.info),
        )



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
SchoolTool calendar views.
"""

import datetime

from zope.component import subscribers
from zope.browserpage.viewpagetemplatefile import ViewPageTemplateFile
from zope.i18n import translate
from schooltool.app.browser import ViewPreferences, same
from schooltool.app.browser.interfaces import ICalendarProvider
from schooltool.calendar.utils import parse_date, week_start
from schooltool.common import SchoolToolMessage as _
from schooltool.app.browser.report import ReportPDFView


SANS = 'Arial_Normal'
SANS_OBLIQUE = 'Arial_Italic'
SANS_BOLD = 'Arial_Bold'
SERIF = 'Times_New_Roman'


class DailyPDFCalendarView(ReportPDFView):
    """The daily view of a calendar in PDF."""

    template=ViewPageTemplateFile('templates/cal_rml.pt')

    title_template = _("Daily calendar for %s")
    subtitle = u""

    @property
    def owner(self):
        return self.context.__parent__.title

    @property
    def title(self):
        return translate(
            self.title_template, context=self.request) % self.owner

    def getDate(self):
        if 'date' in self.request:
            return parse_date(self.request['date'])
        else:
            return datetime.date.today()

    def dayTitle(self, date):
        from schooltool.app.browser.cal import day_of_week_names
        day_of_week_msgid = day_of_week_names[date.weekday()]
        day_of_week = translate(day_of_week_msgid, context=self.request)
        return "%s, %s" % (date.isoformat(), day_of_week)


    def getCalendars(self):
        """Get a list of calendars to display."""
        providers = subscribers((self.context, self.request), ICalendarProvider)

        coloured_calendars = []
        for provider in providers:
            coloured_calendars += provider.getCalendars()

        calendars = [calendar for (calendar, color1, color2)
                     in coloured_calendars]
        return calendars

    def tables(self):
        return [self.buildDayTable(self.getDate())]

    def eventTags(self, event):
        tags = []
        if event.recurrence:
            tags.append(translate(_("recurrent"), context=self.request))
        if (not same(event.__parent__, self.context)
            and event.__parent__ is not None):
            # We have an event from an overlaid calendar.
            tag = translate(_('from the calendar of ${calendar_owner}',
                              mapping={'calendar_owner':
                                       event.__parent__.__parent__.title}),
                            context=self.request)
            # We assume that event.__parent__ is a Calendar which belongs to
            # an object with a title.
            tags.append(tag)
        return tags

    def buildDayTable(self, date):
        """Return the table containing info about events that day."""
        events = self.dayEvents(date)

        rows = []
        for event in events:
            if event.allday:
                time_text = translate(_("all day"), context=self.request)
            else:
                start = event.dtstart.astimezone(self.getTimezone())
                dtend = start + event.duration
                time_text = "%s-%s" % (start.strftime('%H:%M'),
                                            dtend.strftime('%H:%M'))

            row = {
                'title': event.title,
                'description': event.description,
                'location': event.location,
                'time': time_text,
                'resources': ', '.join([
                    resource.title for resource in event.resources]),
                'tags': ', '.join(self.eventTags(event)),
                }

            rows.append(row)

        return {
            'title': self.dayTitle(date),
            'rows': rows,
            }

    def dayEvents(self, date):
        """Return a list of events that should be shown.

        All-day events are placed in front.
        """
        allday_events = []
        events = []
        tz = self.getTimezone()
        start = tz.localize(datetime.datetime.combine(date, datetime.time(0)))
        end = start + datetime.timedelta(days=1)

        for calendar in self.getCalendars():
            for event in calendar.expand(start, end):
                if (same(event.__parent__, self.context)
                      and not same(calendar, self.context)):
                    # We may have overlaid resource booking events appearing
                    # twice (once for self.context and another time for the
                    # other calendar).  We can recognize such dupes by
                    # checking that their __parent__ does not match the
                    # calendar they are coming from.
                    continue
                if event.allday:
                    allday_events.append(event)
                else:
                    events.append(event)

        allday_events.sort()
        events.sort()
        return allday_events + events

    def getTimezone(self):
        """Return the timezone for the PDF report."""
        prefs = ViewPreferences(self.request)
        return prefs.timezone


class WeeklyPDFCalendarView(DailyPDFCalendarView):

    title_template = _("Weekly calendar for %s")

    @property
    def subtitle(self):
        date = self.getDate()
        year, week = date.isocalendar()[:2]
        start = week_start(date, 0) # TODO: first_day_of_week
        end = (start + datetime.timedelta(weeks=1) -
               datetime.timedelta(days=1))
        template = translate(_("Week %d (%s - %s), %d"), context=self.request)
        return template % (week, start, end, year)

    def tables(self):
        start = week_start(self.getDate(), 0) # TODO: first_day_of_week
        return [self.buildDayTable(start + datetime.timedelta(days=weekday))
                for weekday in range(7)]


class MonthlyPDFCalendarView(WeeklyPDFCalendarView):

    title_template = _("Monthly calendar for %s")

    @property
    def subtitle(self):
        from schooltool.app.browser.cal import month_names
        date = self.getDate()
        month_name = translate(month_names[date.month], context=self.request)
        return "%s, %d" % (month_name, date.year)

    def tables(self):
        date = self.getDate()
        day = datetime.date(date.year, date.month, 1)
        tables = []
        while day.month == date.month:
            tables.append(self.buildDayTable(day))
            day = day + datetime.timedelta(days=1)
        return tables



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2010 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Basic stubs for testing SchoolTool.
"""

from persistent.interfaces import IPersistent

from zope.interface import implements
from zope.component import adapts
from zope.keyreference.interfaces import IKeyReference


class KeyReferenceStub(object):
    implements(IKeyReference)
    adapts(IPersistent)

    key_type_id = 'schooltool.testing.stubs.KeyReferenceStub'

    def __init__(self, obj):
        self.object = obj
        self.uid = id(obj)

    def __call__(self):
        return self.object

    def __hash__(self):
        return hash(self.uid)

    def __cmp__(self, other):
        if self.key_type_id != other.key_type_id:
            return cmp(self.key_type_id, other.key_type_id)
        return cmp(self.uid, other.uid)



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2011 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Meta directives and their handlers.
"""
import os

import zope.component.zcml
import zope.configuration.fields
from zope.security.checker import defineChecker, Checker, CheckerPublic
from zope.browserpage import ViewPageTemplateFile
from zope.interface import Interface
from zope.configuration.exceptions import ConfigurationError
from zope.component import zcml
from zope.viewlet.metadirectives import ITemplatedContentProvider

from schooltool.skin.flourish.content import ContentProvider
from schooltool.skin.flourish import interfaces


TEMPLATE_CONTENT_TYPES = {
    'xml': 'text/xml',
    'html': 'text/html',
}


class TemplatePath(zope.configuration.fields.Path):

    def fromUnicode(self, u):
        parts = u.split(':')
        content_type = None
        if len(parts) > 1:
            content_type_part = parts[0].strip().lower()
            if content_type_part in TEMPLATE_CONTENT_TYPES:
                content_type = TEMPLATE_CONTENT_TYPES.get(content_type_part)
                parts.pop(0)
        path = super(TemplatePath, self).fromUnicode(':'.join(parts))
        return content_type, path


class IContentDirective(ITemplatedContentProvider):
    """Define the SchoolTool content provider."""

    template = TemplatePath(
        title=u"Content-generating template.",
        required=False)

    update = zope.configuration.fields.PythonIdentifier(
        title=u"The name of the view attribute implementing content update.",
        required=False,
        )

    render = zope.configuration.fields.PythonIdentifier(
        title=u"The name of the view attribute that renders the content.",
        required=False,
        )

    provides = zope.configuration.fields.GlobalInterface(
        title=u"Interface the component provides",
        required=False,
        default=interfaces.IContentProvider,
        )

    title = zope.configuration.fields.MessageID(
        title=u"Title of this page",
        required=False,
        )


IContentDirective.setTaggedValue('keyword_arguments', True)


class IContentFactoryDirective(zope.component.zcml.IAdapterDirective):

    for_ = zope.configuration.fields.GlobalObject(
        title=u"The interface or class this view is for.",
        required=False,
        default=Interface,
        )

    factory = zope.configuration.fields.GlobalObject(
        title=u"The adapter factory.",
        required=True,
        )

    provides = zope.configuration.fields.GlobalInterface(
        title=u"Interface the component provides",
        required=False,
        default=interfaces.IContentProvider,
        )

    view = zope.configuration.fields.GlobalObject(
        title=u"The view the content provider is registered for.",
        description=(u"The view can either be an interface or a class. By "
                     "default the provider is registered for all views, "
                     "the most common case."),
        required=False,
        default=interfaces.IPageBase,
        )

    layer = zope.configuration.fields.GlobalInterface(
        title=(u"The layer the view is in."),
        required=False,
        default=interfaces.IFlourishLayer,
        )


def handle_security(class_, permission, interfaces, attributes):
    required = set(attributes)
    for ifc in interfaces:
        required.update(set(ifc))

    if permission == 'zope.Public':
        permission = CheckerPublic

    defineChecker(class_, Checker(dict.fromkeys(required, permission)))


def handle_interfaces(_context, interfaces):
    for ifc in interfaces:
        if ifc is not None:
            zope.component.zcml.interface(_context, ifc)


def template_specs(template_dict, content_type=None):
    content_type = TEMPLATE_CONTENT_TYPES.get(content_type, content_type)
    result = dict(template_dict)
    for var in result:
        if (result[var] is not None and
            result[var][0] is None):
            result[var] = (content_type, result[var][1])
    return result


_undefined = object()

def update_specs(template_dict, target):
    types = []
    if issubclass(target, Interface):
        types.append(
            target.queryTaggedValue('flourish.template_content_type', _undefined))
    implemented = getattr(target, '__implemented__', None)
    if implemented:
        for ifc in implemented.interfaces():
            types.append(
                ifc.queryTaggedValue('flourish.template_content_type', _undefined))
    types = filter(lambda p: p is not _undefined, types)
    if types:
        return template_specs(template_dict, types[0])
    return dict(template_dict)


def subclass_content(class_, name,
                     forward_call_dict,
                     template_dict, class_dict):
    class_dict = dict(class_dict)
    class_dict['__name__'] = name
    for attr, template_spec in template_dict.items():
        if not template_spec:
            continue
        content_type, template = template_spec
        if not template:
            continue
        class_dict[attr] = ViewPageTemplateFile(
            template, content_type=content_type)
    classname = (u'%s_%s' % (class_.__name__, name)).encode('ASCII')
    new_class = type(classname, (class_, ), class_dict)
    for attr, base_attr in forward_call_dict.items():
        if attr != base_attr:
            method = getattr(new_class, base_attr)
            setattr(new_class, attr, lambda *a, **kw: method(*a, **kw))
    return new_class


def contentDirective(
    _context, name, permission,
    for_=Interface, layer=interfaces.IFlourishLayer, view=interfaces.IPageBase,
    class_=ContentProvider, template=None,
    update='update', render='render',
    allowed_interface=(), allowed_attributes=(),
    **kwargs):

    if not interfaces.IContentProvider.implementedBy(class_):
        class_ = type(class_.__name__, (class_, ContentProvider), {})

    allowed_interface = (tuple(allowed_interface) +
                         (interfaces.IContentProvider, ))

    if (render == 'render' and
        class_.render == ContentProvider.render):
        if template:
            render = 'template'
        else:
            raise ConfigurationError(
                "When template and render not specified, "
                "class must implement 'render' method")

    templates = update_specs({'template': template}, view)
    class_ = subclass_content(
        class_, name,
        {'update': update, 'render': render},
        templates, kwargs)

    handle_interfaces(_context, (for_, view))
    handle_interfaces(_context, allowed_interface)

    handle_security(class_, permission,
                    allowed_interface, allowed_attributes)

    _context.action(
        discriminator=('schooltool.skin.flourish.content',
                       for_, layer, view, name),
        callable=zope.component.zcml.handler,
        args=('registerAdapter',
              class_,
              (for_, layer, view),
              interfaces.IContentProvider,
              name,
              _context.info),
        )


def contentFactory(
    _context, factory, name='', permission=None,
    for_=Interface, layer=interfaces.IFlourishLayer, view=interfaces.IPageBase,
    provides=interfaces.IContentProvider,
    trusted=True, locate=False):

    wrapper = lambda c, r, v: factory(c, r, v, name)
    objects = [for_, layer, view]
    zope.component.zcml.adapter(
        _context, [wrapper],
        provides=provides,
        for_=objects, permission=permission,
        name=name,
        trusted=trusted, locate=locate)



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
SchoolTool sample data generator interfaces
"""
import zope.interface
import zope.schema


class ISampleDataPlugin(zope.interface.Interface):
    """A plugin that generates some sample data.

    These plugins may depend on other sample data plugins.  Say,
    calendar event generators depend on person generators.  All
    plugins have unique names and other plugins reference their
    dependencies by these names.
    """

    name = zope.schema.Id(title=u"Sample data generator name")

    dependencies = zope.schema.List(
        title=u"A list of dependenies",
        value_type=zope.schema.Id(title=u"Sample data generator name"),
        description=u"""
        A list of names of sample data generators this one depends on.
        """)

    def generate(app, seed=None):
        """Generate sample data of this plugin.

        This method assumes the sample data this plugin depends on has
        been created.
        """


class CyclicDependencyError(ValueError):
    """Cyclic dependency of sample data plugins"""



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
SchoolTool internal error view.
"""

import sys
import linecache
import cgi

from zope.app.applicationcontrol.interfaces import IApplicationControl
from zope.app.applicationcontrol.interfaces import IRuntimeInfo
from zope.cachedescriptors.property import Lazy
from zope.interface import implements
from zope.publisher.browser import BrowserView
from zope.browser.interfaces import ISystemErrorView

from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.common import HTMLToText, find_launchpad_project


class ErrorView(BrowserView):
    """A view for server error messages."""

    implements(ISystemErrorView)

    def isSystemError(self):
        return True

    def __call__(self):
        self.request.response.setStatus(500)
        return self.index()

    def plaintext_traceback(self):
        # Not proud of this :)
        parser = HTMLToText()
        parser.feed(unicode(self.traceback))
        plaintext = parser.get_data()
        return plaintext

    @Lazy
    def bugreport_package(self):
        etype, value, tb = sys.exc_info()
        package = find_launchpad_project(tb)
        return package

    @Lazy
    def bugreport_url(self):
        package = self.bugreport_package
        url = "https://bugs.launchpad.net/%s/+filebug" % package
        return url

    @Lazy
    def traceback(self):
        # Note: this function exposes filenames to the web user; some may
        # consider it a security risk.
        etype, value, tb = sys.exc_info()
        lines = []
        w = lines.append
        q = lambda s: cgi.escape(str(s), True)
        for filename, lineno, method, locals in extract_tb(tb):
            w('File "<span class="filename">%s</span>",'
                    ' line <span class="lineno">%s</span>,'
                    ' in <span class="method">%s</span>\n'
              % (q(filename), q(lineno), q(method)))
            w('  <span class="source">%s</span>\n'
              % q(linecache.getline(filename, lineno).strip()))
            self._extra_info(w, dict(locals))
        return "".join(lines)

    def _extra_info(self, w, locals):
        # TODO: unit tests would be nice
        q = lambda s: cgi.escape(str(s), True)
        if '__traceback_info__' in locals:
            tb_info = locals['__traceback_info__']
            w('Extra information: %s\n' % q(repr(tb_info)))
        if '__traceback_supplement__' in locals:
            tb_supplement = locals['__traceback_supplement__']
            tb_supplement = tb_supplement[0](*tb_supplement[1:])
            from zope.pagetemplate.pagetemplate import \
                    PageTemplateTracebackSupplement
            from zope.tales.tales import TALESTracebackSupplement
            if isinstance(tb_supplement, PageTemplateTracebackSupplement):
                source_file = tb_supplement.manageable_object.pt_source_file()
                if source_file:
                    w('Template "<span class="filename">%s</span>"\n'
                      % q(source_file))
            elif isinstance(tb_supplement, TALESTracebackSupplement):
                w('Template "<span class="filename">%s</span>",'
                  ' line <span class="lineno">%s</span>,'
                  ' column <span class="column">%s</span>\n'
                  % (q(tb_supplement.source_url), q(tb_supplement.line),
                     q(tb_supplement.column)))
                w('  Expression: <span class="expr">%s</span>\n'
                  % q(tb_supplement.expression))
            else:
                w('__traceback_supplement__ = %s\n'
                  % q(repr(tb_supplement)))

    @Lazy
    def runtime(self):
        app = ISchoolToolApplication(None)
        application_control = IApplicationControl(app, None)
        if application_control is None:
            return None
        try:
            ri = IRuntimeInfo(application_control)
        except TypeError:
            return None
        runtime = (
            'Python %s' % ri.getPythonVersion(),
            ri.getSystemPlatform(),
            'Filesystem encoding %s, preferred %s' % (
                ri.getFileSystemEncoding(), ri.getPreferredEncoding()
                ),
            )
        return runtime


def extract_tb(tb, limit=None):
    """Improved version of traceback.extract_tb.

    Includes a dict with locals in every stack frame instead of the line.
    """
    # TODO: unit tests would be nice
    list = []
    while tb is not None and (limit is None or len(list) < limit):
        frame = tb.tb_frame
        code = frame.f_code
        name = code.co_name
        filename = code.co_filename
        lineno = tb.tb_lineno
        locals = frame.f_locals
        list.append((filename, lineno, name, locals))
        tb = tb.tb_next
    return list
